# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
"""
NAM: ReaMBFits.py (file)
DES: Provides classes and methodes for the low level access to MBFits datasets.
     The Module Interface consists of
     - Function isDataset
     - Function importDataset
     - Function openDataset
     - Class Dataset
     - Class Table
     - Class Keyword
     - Class Column
     - Class ColumnInfo
     - Class MBFitsError
     Only these functions and classes should be used by clients!

     The rest of the module contains classes that implement the module's
     functionality. These classes should not be used directly from outside the module,
     but only through the module's interface!
"""
# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------

import os
import numpy as np
import cfitsio

# ---------------------------------------------------------------------
# ---- Module Interface -----------------------------------------------
# ---------------------------------------------------------------------

# ---------------------------------------------------------------------
# ---- Function isDataset ---------------------------------------------
# ---------------------------------------------------------------------


def isDataset(datasetName):
    """
    NAM: isDataset (Function)
    DES: Checks if datasetName corresponds to a dataset
    INP: datasetName (str) : The full path to the directory or file to be checked
    OUT: (long)            : 1, if datasetName is a dataset, 0 else
    """
    if os.path.isdir(datasetName):
        return os.path.isfile(datasetName + "/GROUPING.fits") or \
            os.path.isfile(datasetName + "/GROUPING.fits.gz")
    else:
        return (datasetName[-5:] == ".fits" or datasetName[-8:] == ".fits.gz")

# ---------------------------------------------------------------------
# ---- Function importDataset -----------------------------------------
# ---------------------------------------------------------------------


def importDataset(datasetName, iomode=0):
    """
    NAM: importDataset (Function)
    DES: Creates a new Dataset object and imports it from disk files specified by
         datasetName. The return Dataset object is opened for reading (iomode=0) or
         reading and writing (iomode=1); however, the Dataset's tables are not opened.
    INP: datasetName (str): The full path of the directory or file of the dataset
         iomode     (long): 0: the returned Dataset is open for reading
                            1: the returned Dataset is open for reading and writing
    OUT: (Dataset)        : The imported Dataset object
    """
    datasetImpl = _Dataset(datasetName)
    if os.path.isdir(datasetName):
        if os.path.isfile(datasetName + "/GROUPING.fits"):
            datasetImpl.imprt(datasetName + "/GROUPING.fits", iomode)
        elif os.path.isfile(datasetName + "/GROUPING.fits.gz"):
            datasetImpl.imprt(datasetName + "/GROUPING.fits.gz", iomode)
        else:
            raise MBFitsError("Direcotry %s no Dataset" % datasetName)
    else:
        datasetImpl.imprt(datasetName, iomode)

    return Dataset(datasetImpl)

# ---------------------------------------------------------------------
# ---- Function createDataset -----------------------------------------
# ---------------------------------------------------------------------


def createDataset(datasetName, filename, keywords, groupName):
    """
    NAM: createDataset (Function)
    DES: Creates a new Dataset object without Tables. The created dataset
         is opened for reading and writing.
    INP: datasetName       (str): The full path of the directory or file of the dataset
         filename          (str): The full name of the file containig the primary header;
                                  may be identical to datasetName
         keywords (Keyword list): List of Keyword objects to be added to the dataset
         groupName         (str): Name of the hierarchical group defined in the dataset's
                                  Grouping Table. If None, no Grouping Table is created.
    OUT: (Dataset)              : The created Dataset object
    """
    datasetImpl = _Dataset(datasetName)
    datasetImpl.construct(filename, keywords, groupName)
    return Dataset(datasetImpl)

# ---------------------------------------------------------------------
# ---- Class Dataset --------------------------------------------------
# ---------------------------------------------------------------------


class Dataset:

    """
    NAM: Dataset (Class)
    DES: Represents a MBFits dataset.
    """

    def __init__(self, implementation):
        """
        NAM: Dataset.__init__ (Method)
        DES: Constructor of Class Dataset
             Do not use this constructor from outside the module; instead, create
             Dataset objects via the functions importDataset or createDataset
        """
        self._implementation = implementation

    # -----------------------------------------------------------------

    def isOpen(self):
        """
        NAM: Dataset.isOpen (Method)
        DES: Check if Dataset is open for reading
        OUT: (long): 1, if Dataset is open for reading, 0 otherwise
        """
        if not self._implementation:
            return 0
        return self._implementation.isOpen()

    # -----------------------------------------------------------------

    def isWriteOpen(self):
        """
        NAM: Dataset.isWriteOpen (Method)
        DES: Check if Dataset is open for reading and writing
        OUT: (long): 1, if Dataset is open for reading and writing, 0 otherwise
        """
        if not self._implementation:
            return 0
        return self._implementation.isWriteOpen()

    # -----------------------------------------------------------------

    def open(self, iomode=0):
        """
        NAM: Dataset.open (Method)
        DES: Open the Dataset for reading or reading plus writing.
             A Dataset must be open to perform Dataset.getTables.
             A Dataset must be open for reading and writing in order to
             perform Dataste.addTable.
             Tables can only be opened if the corresponding Dataset is open.
             Note that this method does not open the Dataset's Tables.
        INP: iomode (long): 0: open Dataset for reading
                            1: open Dataset for reading and writing
                            Note that the Dataset's Tables can only be opened
                            for reading and writing if the Dataset was opened
                            with iomode=1
        """
        if self._implementation:
            self._implementation.open(iomode=iomode)

    # -----------------------------------------------------------------

    def close(self):
        """
        NAM: Dataset.close (Method)
        DES: Close the Dataset and all its Tables.
        """
        if self._implementation:
            self._implementation.close()

    # -----------------------------------------------------------------

    def exprt(self):
        """
        NAM: Dataset.exprt (Method)
        DES: Export the Dataset to disk.
             Export writes the keywords of the Primary Header and of the
             Grouping Table to disk; in addition, it creates the necessary
             columns of the Grouping Table if necessary.
             Note that this method does not export the Dataset's Tables.
        """
        if self._implementation:
            self._implementation.exprt()

    # -----------------------------------------------------------------

    def getName(self):
        """
        NAM: Dataset.getName (Method)
        DES: Returns the Dataset's name
        OUT: (str): The full path of the directory or file of the Dataset
        """
        if not self._implementation:
            return None
        return self._implementation.getName()

    # -----------------------------------------------------------------

    def getSize(self):
        """
        NAM: Dataset.getSize (Method)
        DES: Returns the sum of the sizes of all files that make up the Dataset.
        OUT: (long): The size (in byte) of the Dataset.
        """
        if not self._implementation:
            return 0
        return self._implementation.getSize()

    # -----------------------------------------------------------------

    def getKeywordNames(self):
        """
        NAM: Dataset.getKeywordNames (Method)
        DES: Returns the names of the Keywords in the Dataset's Primary Header
             in the correct order
        OUT: (str list): The names of Keywords in the Dataset's Primary Header
        """
        if not self._implementation:
            return None
        return self._implementation.getKeywordNames()

    # -----------------------------------------------------------------

    def getKeyword(self, keyname):
        """
        NAM: Dataset.getKeyword (Method)
        DES: Returns a Keyword from the Dataset's Primary Header
        INP: keyname (str): The keyname of the requested Keyword
        OUT: (Keyword)    : The specified Keyword object from the Primary Header
        """
        if not self._implementation:
            return None
        return self._implementation.getKeyword(keyname)

    # -----------------------------------------------------------------

    def getTables(self, **keywords):
        """
        NAM: Dataset.getTables (Method)
        DES: Returns Tables from the Dataset.
             getTables(keyname1=value1, keyname2=value2,...) returns all Tables with
             matching keywords, so that for each returned Table t
             t.getKeyword("keyword1")==value1 and t.getKeyword("keyword2")==value2 and ...
             is true.
             Without argument, getTables() returns all Tables.
        INP: **keywords  : Arbitrary number of keyword arguments of type
                           keyname=keyvalue.
        OUT: (Table list): The specified Table objects
        """
        if not self._implementation:
            return None
        tableImpls = self._implementation.getTables(**keywords)
        tables = []
        for tableImpl in tableImpls:
            if tableImpl:
                tables.append(Table(tableImpl))
        return tables

    # -----------------------------------------------------------------

    def addTable(self, filename=None, keywords=[], colinfos=[]):
        """
        NAM: Dataset.addTable (Method)
        DES: Adds a Table to the Dataset.
             The added Table is created in the file specified by filename. If necessary,
             the file is created (including dirctories along its path). If no filename is
             given, the new table is created in the file that holds the Dataset's
             Primary Header.
             The new Table will have Keywords and Columns as specified by the arguments keywords
             and colinfos. Note that in the course of creation of the Table, some additional
             Keywords may be created automatically; these Keywords can be accessed via
             Table.getKeyword exactly as the Keywords specified explicitely in
             Dataset.addTable.
             During execution of Dataset.addTable, the new Table is added to the Dataset's
             Grouping Table, if this exists. The new Table will be open for reading and
             writing, and is exported during execution of Dataset.addTable.
             The Dataset must be open for reading and writing.
        INP: filename (str)            : The full path of the file in which the new Table is created.
                                         If None, the new Table is created in the file with the
                                         Dataset's Primary Header.
             keywords (Keyword list)   : Keywords to be added to the new Table.
             colinfos (ColumnInfo list): Colinfo objects that describe the columns of the
                                         new Table.
        OUT: (Table)                   : The newly created Table
        """
        if not self._implementation:
            return None
        tableImpl = self._implementation.addTable(filename, keywords, colinfos)
        if not tableImpl:
            return None
        table = Table(tableImpl)
        return table

# ---------------------------------------------------------------------
# ---- Class Table ----------------------------------------------------
# ---------------------------------------------------------------------


class Table:

    """
    NAM: Table (Class)
    DES: Represents a table of a MBFits dataset.
    """

    def __init__(self, implementation):
        """
        NAM: Table.__init__ (Method)
        DES: Constructor of Class Table
             Do not use this constructor from outside the module; instead, create
             Table objects via the methods Dataset.getTables and .addTable
        """
        self._implementation = implementation

    # -----------------------------------------------------------------

    def isOpen(self):
        """
        NAM: Table.isOpen (Method)
        DES: Check if Table is open for reading
        OUT: (long): 1, if Table is open for reading, 0 otherwise
        """
        if not self._implementation:
            return 0
        return self._implementation.isOpen()

    # -----------------------------------------------------------------

    def isWriteOpen(self):
        """
        NAM: Table.isWriteOpen (Method)
        DES: Check if Table is open for reading and writing
        OUT: (long): 1, if Table is open for reading and writing, 0 otherwise
        """
        if not self._implementation:
            return 0
        return self._implementation.isWriteOpen()

    # -----------------------------------------------------------------

    def open(self, iomode=0):
        """
        NAM: Table.open (Method)
        DES: Open the Table for reading or reading plus writing.
             A Table must be open to perform Table.close, .getColumn,
             .getNumRows, and .setSelection.
             A Table must be open for reading and writing in order to
             perform Table.exprt.
             A Table can only be opened if the corresponding Dataset is open.
             A Table can only be opened for reading and writing if the corresponding
             Dataset is open for reading and writing.
        INP: iomode (long): 0: open Table for reading
                            1: open Table for reading and writing
                            Note that a Table can only be opened
                            for reading and writing if the Dataset was opened
                            with iomode=1
        """
        if self._implementation:
            self._implementation.open(iomode=iomode)

    # -----------------------------------------------------------------

    def close(self):
        """
        NAM: Table.close (Method)
        DES: Close the Table.
        """
        if self._implementation:
            self._implementation.close()

    # -----------------------------------------------------------------

    def exprt(self):
        """
        NAM: Table.exprt (Method)
        DES: Export the Table to disk.
             Export writes the Table's keywords to disk; in addition, it creates
             the Table's columns if necessary.
             The Table must be open for reading and writing to call this method.
        """
        if self._implementation:
            self._implementation.exprt()

    # -----------------------------------------------------------------

    def getKeywordNames(self):
        """
        NAM: Table.getKeywordNames (Method)
        DES: Returns the names of the Keywords in the correct order
        OUT: (str list): The names of Keywords in the Table
        """
        if not self._implementation:
            return None
        return self._implementation.getKeywordNames()

    # -----------------------------------------------------------------

    def getKeyword(self, keyname):
        """
        NAM: Table.getKeyword (Method)
        DES: Returns a Keyword from the Table
        INP: keyname (str): The keyname of the requested Keyword
        OUT: (Keyword)    : The specified Keyword object
        """
        if not self._implementation:
            return None
        return self._implementation.getKeyword(keyname)

    # -----------------------------------------------------------------

    def getNumColumns(self):
        """
        NAM: Table.getNumColumns (Method)
        DES: Returns the number of Columns
        OUT: (long): The number of Columns in the Table
        """
        if not self._implementation:
            return None
        return self._implementation.getNumColumns()

    # -----------------------------------------------------------------

    def getColumnNames(self):
        """
        NAM: Table.getColumnNames (Method)
        DES: Returns the names of the Columns in the correct order.
        OUT: (str list): The names of Columns in the Table
        """
        if not self._implementation:
            return None
        return self._implementation.getColumnNames()

    # -----------------------------------------------------------------

    def getColumn(self, colName):
        """
        NAM: Table.getColumn (Method)
        DES: Returns the Column object with name colName.
             The Table must be open to call this method.
        INP: colName (str): The name of the requesetd Column
        OUT: (Column):      The Column object
        """
        if not self._implementation:
            return None
        columnImpl = self._implementation.getColumn(colName)
        if not columnImpl:
            return None
        column = Column(columnImpl)
        return column

    # -----------------------------------------------------------------

    def getNumRows(self):
        """
        NAM: Table.getNumRows (Method)
        DES: Returns the number of rows in the Table's columns.
             The Table must be open to call this method.
        OUT: (long): The number of rows in the Table's columns.
        """
        if not self._implementation:
            return None
        return self._implementation.getNumRows()

    # -----------------------------------------------------------------

    def getOptimalRowsize(self):
        """
        NAM: Table.getOptimalRowsize (Method)
        DES: Returns the optimal number of rows to be read or written in a single call
             of Column.read or .write for this Table.
             This number depends on the number of files that are open at the time of
             reading or writing. Hence for optimal performance, call this function after
             opening or closing a Table.
        OUT: (long): The optimal number of rows for reading and writing.
        """
        if not self._implementation:
            return None
        return self._implementation.getOptimalRowsize()

    # -----------------------------------------------------------------

    def setSelection(self, expression, firstRow=1, numRows=0):
        """
        NAM: Table.setSelection (Method)
        DES: Sets a selection to the Table.
             If a selection is set to a Table, subsequent calls of
             Column.read will return only those rows, where the specified boolean
             expression is true.
             For a full description of expression, firstRow and numRows, see
             the documentation of the CFITSIO routine fits_find rows.
             The Table must be open to call this method.
        INP: expression (str): Boolean expression which defines the selection
             firstRow (long):  First Row to which the selection applies.
             numRows(long):    Number of rows, to which the selection applies;
                               if None or 0, the selection applies to all rows
                               after firstRow
        """
        if self._implementation:
            self._implementation.setSelection(expression, firstRow, numRows)

    # -----------------------------------------------------------------

    def clearSelection(self):
        """
        NAM: Table.clearSelection (Method)
        DES: Clears a selection
        """
        if self._implementation:
            self._implementation.clearSelection()

    # -----------------------------------------------------------------

    def hasSelection(self):
        """
        NAM: Table.hasSelection (Method)
        DES: Check if a Table has a selection
        OUT: 1 if true; 0 else
        """
        if not self._implementation:
            return None
        return self._implementation.hasSelection()


# ---------------------------------------------------------------------
# ---- Class Keyword --------------------------------------------------
# ---------------------------------------------------------------------

class Keyword:

    """
    NAM: Keyword (Class)
    DES: Represents a keyword of a MBFits dataset
    """

    def __init__(
        self,
        name='',
     datatype='',
     value=None,
     format='',
     comment='',
     unit=''):
        """
        NAM: Keyword.__init__
        DES: Constructor of class Keyword
        INP: name (str):     The Keyword's name
             datatype (str): The Keywords datatype coded in the spirit of CFITSIO:
                             'A' - String
                             'L' - Logical
                             'I' - Integer
                             'J' - Long
                             'E' - Float
                             'D' - Double
             value:          The Keyword's value of the appropriate datatype.
             format (long):  The Keywords format used when writing the Keyword to disk.
                             For datatype
                             'A': The maximum length of the string
                             'E': The number of decimal digits
                             'D': The number of decimal digits
                             For all other datatypes unused.
             comment (str):  The comment of the Keyword; Note that it may or may not contain
                             the string describing the unit.
             unit (str):     The unit of the keyword.
        """
        self._name = name
        self._datatype = datatype
        self._value = value
        self._format = format
        self._comment = comment
        self._unit = unit

    # -----------------------------------------------------------------

    def __str__(self):
        return str((self._name,
                    self._value,
                    self._unit,
                    self._format,
                    self._comment,
                    self._datatype))

    # -----------------------------------------------------------------

    def getName(self):
        """
        NAM: Keyword.getName
        DES: Retuns the keyword's name
        """
        return self._name

    # -----------------------------------------------------------------

    def getDatatype(self):
        """
        NAM: Keyword.getDatatype
        DES: Retuns the keyword's datatype
        """
        return self._datatype

    # -----------------------------------------------------------------

    def getValue(self):
        """
        NAM: Keyword.getValue
        DES: Retuns the keyword's value
        """
        return self._value

    # -----------------------------------------------------------------

    def getFormat(self):
        """
        NAM: Keyword.getFormat
        DES: Retuns the keyword's format
        """
        return self._format

    # -----------------------------------------------------------------

    def getComment(self):
        """
        NAM: Keyword.getComment
        DES: Retuns the keyword's comment
        """
        return self._comment

    # -----------------------------------------------------------------

    def getUnit(self):
        """
        NAM: Keyword.getUnit
        DES: Retuns the keyword's unit
        """
        return self._unit

    # -----------------------------------------------------------------

    def setValue(self, value):
        """
        NAM: Keyword.setValue
        DES: Sets the keyword's value
        """
        self._value = value

    # -----------------------------------------------------------------

    def setFormat(self, format):
        """
        NAM: Keyword.setFormat
        DES: Sets the keyword's format
        """
        self._format = format

    # -----------------------------------------------------------------

    def setComment(self, comment):
        """
        NAM: Keyword.setComment
        DES: Sets the keyword's comment
        """
        self._comment = comment

    # -----------------------------------------------------------------

    def setUnit(self, unit):
        """
        NAM: Keyword.setUnit
        DES: Sets the keyword's unit
        """
        self._unit = unit


# ---------------------------------------------------------------------
# ---- Class Column ---------------------------------------------------
# ---------------------------------------------------------------------

class Column:

    """
    NAM: Column (Class)
    DES: Represents a column of a MBFits dataset
    """

    def __init__(self, implementation):
        """
        NAM: Column.__init__ (Method)
        DES: Constructor of Class Column
             Do not use this constructor from outside the module; instead, create
             Column objects via the methods Table.getColumn
        """
        self._implementation = implementation

    # -----------------------------------------------------------------

    def getColnum(self):
        """
        NAM: Column.getColnum (Method)
        DES: Returns the column number. The first number of a Table has column number 1
        OUT: (long): The column number
        """
        if not self._implementation:
            return None
        return self._implementation.getColnum()

    # -----------------------------------------------------------------

    def getName(self):
        """
        NAM: Column.getName (Method)
        DES: Returns the column's name.
        OUT: (str): The column's nane
        """
        if not self._implementation:
            return None
        return self._implementation.getName()

    # -----------------------------------------------------------------

    def getDatatype(self):
        """
        NAM: Column.getDatatype (Method)
        DES: Returns the column's datatype.
             See Class ColumnInfo for further documentation.
        OUT: (str): The column's datatype
        """
        if not self._implementation:
            return None
        return self._implementation.getDatatype()

    # -----------------------------------------------------------------

    def getRepeat(self):
        """
        NAM: Column.getRepeat (Method)
        DES: Returns the column's repeat count.
             See Class ColumnInfo for further documentation.
        OUT: (str): The column's repeat count
        """
        if not self._implementation:
            return None
        return self._implementation.getRepeat()

    # -----------------------------------------------------------------

    def getDim(self):
        """
        NAM: Column.getDim (Method)
        DES: Returns the column's dimension.
             See Class ColumnInfo for further documentation.
        OUT: (str): The column's dimension
        """
        if not self._implementation:
            return None
        return self._implementation.getDim()

    # -----------------------------------------------------------------

    def getDescription(self):
        """
        NAM: Column.getDescription (Method)
        DES: Returns the column's description.
             See Class ColumnInfo for further documentation.
        OUT: (str): The column's description.
        """
        if not self._implementation:
            return None
        return self._implementation.getDescription()

    # -----------------------------------------------------------------

    def getUnit(self):
        """
        NAM: Column.getUnit (Method)
        DES: Returns the column's unit.
             See Class ColumnInfo for further documentation.
        OUT: (str): The column's unit.
        """
        if not self._implementation:
            return None
        return self._implementation.getUnit()

    # -----------------------------------------------------------------

    def read(self, firstRow=1, numRows=0):
        """
        NAM: Column.read (Method)
        DES: Reads data from a Column.
             Reads numRows rows starting at row firstRow. If numRows=0, all rows
             starting from firstRow are read. Note that row numbers start at 1.
             The datatype of the output depends on the Column's datatype, dimension,
             and repeat count. Whenever possible, a np.array is returned with the
             relation of the column's datatype and the np.array's typecode as follows:
                 'L'                        -> np.int0
                 'B'                        -> np.unit8
                 'I', 'I', 'UI'             -> np.int16
                 'K', 'UK', 'J', 'JJ', 'UJ' -> np.int32
                 'E', 'F'                   -> np.float32
                 'D', 'G'                   -> np.float64
             The shape of the returned Numeric array depends on the column's dimension; in general,
             the rank of the array is 1 higher than the rank of the column's dimension. This
             is also true, if only one row was read.
             For the column's datatype = 'A', a list of strings is returned.
             For variable length arrays, a list of 1-dim np.arrays of the appropriate size
             is returned.
        INP: firstRow (long): The number of the first Row to be read
             numRows (long):  The number of rows to be read. If 0, all rows starting from firstRow
                              are read.
        OUT: See DES
        """
        if not self._implementation:
            return None
        return self._implementation.read(firstRow, numRows)

    # -----------------------------------------------------------------
    def write(self, firstRow, data):
        """
        NAM: Column.write (Method)
        DES: Writes data to a Column.
             Writing starts at row firstRow. Note that row numbers start at 1.
             For the datatype of data the corresponding rules to Column.read apply:
             For columns that hold numeric data of not-variable length, data must be a
             Numeric array. If the Numeric array's typecode does not match the typecode
             as specified in Column write, data is cast to the correct typecode. (Note that
             this introduces a performance penalty!). The shape of the Numeric array need
             not match the dimension of the Column; however, the number of elements in data
             must be correct.
             For columns with varable length numeric data, data must be a list of np.arrays.
             Each element of the list will be written into a separate row. Concerning
             typecode the above said applies.
             For string data (both of fixed length and variable length), data must be a
             list of strings.
        INP: firstRow (long): The number of the first Row to which data is written
             data:            See DES
        """
        if self._implementation:
            self._implementation.write(firstRow, data)

# ---------------------------------------------------------------------
# ---- Class ColumnInfo -----------------------------------------------
# ---------------------------------------------------------------------


class ColumnInfo:

    """
    NAM: ColumnInfo (Class)
    DES: Contains the information that is needed for the creation of a single column
         during the execution of Dataset.addTable.
         The meaning of the data elements is:
         - name:        The name of the column as in Keyword TTYPEn
         - datatype:    The datatype code for the column as specified in Keyword TFORMn.
                        Must not contain the repeat count nor the length for variable
                        length arrays. The repeat count is specccified in repeat, wheras
                        the length is added automatically when the Table is closed.
                        See Column.read and the CFITSIO documentation for valid datatypes.
         - repeat:      The repeat count as specified in Keyword TFORMn. In case of string
                        data: The maximal length of the string.
         - dim:         The dimension as specified in Keyword TDIMn.
         - description: A descriptive text for the column. Stored as comment of the Keyword
                        TTYPEn.
         - unit:        The unit for the column. Stored as part of the comment of the Keyword
                        TTYPEn.
         If both dim and repeat are specified, the repeat count in Keyword TFORMn is evaluated
         from dim. repeat is ignored in this case.
    """

    def __init__(
        self,
        name='',
     datatype='',
     repeat=0,
     dim=[],
     description='',
     unit=''):
        self.name = name
        self.datatype = datatype
        self.repeat = repeat
        self.dim = dim
        self.description = description
        self.unit = unit

# ---------------------------------------------------------------------
# ---- Class MBFitsError ----------------------------------------------
# ---------------------------------------------------------------------


class MBFitsError(Exception):

    """
    NAM: MBFitsError (Class)
    DES: Exception class for exceptions related with module ReaMBFits
    """

    def __init__(self, msg, code=None):
        self.code = code
        self.msg = msg

    # -----------------------------------------------------------------

    def __str__(self):
        return repr(self.msg)


# ---------------------------------------------------------------------
# ---- Implementation Classes -----------------------------------------
# ---------------------------------------------------------------------

# ---------------------------------------------------------------------
# ---- Class _Dataset ----------------------------------------------
# ---------------------------------------------------------------------

class _Dataset:
    # NAM: _Dataset (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, name):
        self._name = name

        self._isopen = 0
        self._iomode = 0

        self._primaryHeader = None
        self._groupingTable = None
        self._tableHdus = []

    # -----------------------------------------------------------------

    def construct(self, filename, keywords=[], groupName=None):
        # Construct a new Dataset with Primary Header and Grouping Table
        # in the specified file.
        # The Primary Header will contain the specified Keywords.
        # The Grouping Table will contain a group with name groupName.
        # If groupName=None, no Grouping Table will be created.

        if self._primaryHeader:
            raise MBFitsError("Dataset %s exists already" % self._name)

        if os.path.exists(filename):
            raise MBFitsError("File %s exists already" % filename)

        self._isopen = 1
        self._iomode = 1

        fptr = self._createFile(filename, isDatasetFile=1)
        self._primaryHeader = _Hdu(fptr, self, iomode=1)
        for keyword in keywords:
            self._primaryHeader.addKeyword(keyword)
        self._primaryHeader.exprt()

        if groupName:
            self._groupingTable = self._createGroupingTable(
                filename, groupName)
            self._groupingTable.exprt()

        # Flush the file and reimport to synchronize keywords:
        status = cfitsio.fits_flush_file(fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        self._primaryHeader.imprt()
        if self._groupingTable:
            self._groupingTable.imprt()

    # -----------------------------------------------------------------

    def isOpen(self):
        return self._isopen

    # -----------------------------------------------------------------

    def isWriteOpen(self):
        return self.isOpen() and self._iomode

    # -----------------------------------------------------------------

    def open(self, iomode=0):
        if self.isOpen():
            raise MBFitsError("Dataset %s allready open" % self._name)

        self._isopen = 1
        self._iomode = iomode

        if not self._primaryHeader:
            raise MBFitsError("Dataset %s has no Primary Header" % self._name)
        self._primaryHeader.open(iomode)

        if self._groupingTable:
            self._groupingTable.open(iomode)

    # -----------------------------------------------------------------

    def close(self):
        if not self.isOpen():
            raise MBFitsError("Dataset %s not open" % self._name)

        for table in self._tableHdus:
            if table.isOpen():
                table.close()

        if self._groupingTable:
            if self._groupingTable.isOpen():
                self._groupingTable.close()

        if self._primaryHeader.isOpen():
            self._primaryHeader.close()

        self._isopen = 0

    # -----------------------------------------------------------------

    def imprt(self, filename, iomode):
        # Import Dataset from disk. filename specifies the file that contains the Primary Header.
        # On exit, the Dataset will be open, while all Tables will be closed.

        if self._primaryHeader:
            raise MBFitsError("Dataset %s exists already" % self._name)

        # Open the specified file:
        status, fptr = cfitsio.fits_open_file(filename, iomode)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        if not fptr:
            raise MBFitsError("Failed to open file %s" % filename)

        self._isopen = 1
        self._iomode = iomode

        status, nHdus = cfitsio.fits_get_num_hdus(fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Walk through the specified file and create _TableHdu-objects:
        for iHdu in np.arange(1, nHdus + 1):
            # Reopen the file to get a new file pointer to be used for this
            # HDU:
            status, fptrHdu = cfitsio.fits_reopen_file(fptr)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)
            if not fptrHdu:
                raise MBFitsError(
                    "Failed to reopen file %s for HDU %d" %
                    (filename, iHdu))

            # Move the new file pointer to the HDU:
            status, cpHduNum = cfitsio.fits_movabs_hdu(fptrHdu, iHdu)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            if iHdu == 1:
                # Primary Header:
                self._primaryHeader = _Hdu(fptrHdu, self, iomode)
                self._primaryHeader.imprt()
                # The primary header is left open:
                # self._primaryHeader.close()
            else:
                # Tables:
                if not _GroupingHdu.isGroupingHdu(fptrHdu):
                    # Create _TableHdu for this Hdu:
                    tableHdu = _TableHdu(fptrHdu, self, iomode)
                    tableHdu.imprt()
                    tableHdu.close()
                    self._tableHdus.append(tableHdu)
                else:
                    if self._groupingTable:
                        raise MBFitsError(
                            "More than 1 GROUPING HDU in file %s" %
                            filename)
                    # Create _GroupingHdu for this HDU:
                    self._groupingTable = _GroupingHdu(fptrHdu, self, iomode)
                    # Create _TableHdus for all HDUs referenced in this
                    # Grouping HDU:
                    memberTableHdus = self._groupingTable.importMemberTableHdus(
                    )
                    self._tableHdus.extend(memberTableHdus)
                    # The Grouping HDU is left open:
                    # self._groupingTable.close()

        # Close the dataset file:
        # Primary Header and Grouping Table remain open!
        status = cfitsio.fits_close_file(fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

    # -----------------------------------------------------------------

    def exprt(self):
        # Export the Dataset, ie. the Primary Header and the Grouping Table.
        # The Dataset's Tables will nopt be exported.

        if not self.isWriteOpen():
            raise MBFitsError("Dataset %s not open to write" % self._name)

        self._primaryHeader.exprt()

        if self._groupingTable:
            self._groupingTable.exprt()

    # -----------------------------------------------------------------

    def getName(self):
        return self._name

    # -----------------------------------------------------------------

    def getFilename(self):
        if self._primaryHeader:
            return self._primaryHeader.getFilename()
        else:
            return None

    # -----------------------------------------------------------------

    def getSize(self):
        size = 0
        if self._primaryHeader:
            filenames = [self._primaryHeader.getFilename()]
            if self._groupingTable:
                filename = self._groupingTable.getFilename()
                if not filename in filenames:
                    filenames.append(filename)
            for tableHdu in self._tableHdus:
                filename = tableHdu.getFilename()
                if not filename in filenames:
                    filenames.append(filename)

            for filename in filenames:
                size += os.path.getsize(filename)

        return size

    # -----------------------------------------------------------------

    def getKeywordNames(self):
        if not self._primaryHeader:
            return None
        return self._primaryHeader.getKeywordNames()

    # -----------------------------------------------------------------

    def getKeyword(self, keyname):
        if not self._primaryHeader:
            return None
        return self._primaryHeader.getKeyword(keyname)

    # -----------------------------------------------------------------

    def getTables(self, **kargs):
        if not self.isOpen():
            raise MBFitsError("Dataset %s not open" % self._name)

        if not kargs:
            # Return all tableHdus if no arguments are passed:
            result = self._tableHdus
        else:
            # Return only matching tableHdus if arguments are passed:
            result = []
            for tableHdu in self._tableHdus:
                keywordsMatch = 1
                for karg in kargs.keys():
                    tableKeyword = tableHdu.getKeyword(karg)
                    if kargs[karg] is not None:
                        #  keyname=value in kargs
                        if tableKeyword is None:
                            # Keyword keyname not present in tableHDU:
                            keywordsMatch = 0
                            break
                        if tableKeyword.getValue() != kargs[karg]:
                            # Keyword keyname present in tableHDU,
                            # but value does not match:
                            keywordsMatch = 0
                            break
                    else:
                        #  keyname=None in kargs
                        if tableKeyword is not None:
                            # Keyword keyname present in tableHDU:
                            keywordsMatch = 0
                            break
                if keywordsMatch:
                    result.append(tableHdu)
        return result

    # -----------------------------------------------------------------

    def addTable(self, filename=None, keywords=[], colinfos=[]):
        # Creates a new Table in the specified file. If filename=None, the Table
        # will be created in the file that contains the Primary Header.
        # The Table will contain the specified Keywords and Columns.
        # The Table will be added to the Grouping Table, if a Grouping Table exists.
        # The Table will be open for reading and writing on exit from this
        # method.

        if not self.isWriteOpen():
            raise MBFitsError("Dataset %s not open to write" % self._name)

        if not filename:
            filename = self._primaryHeader.getFilename()
            fileExists = 1
        else:
            # See if the specified file already exists:
            hdus = [self._primaryHeader, self._groupingTable]
            hdus.extend(self._tableHdus)

            fileExists = 0
            for hdu in hdus:
                if hdu:
                    if hdu.getFilename() == filename:
                        fileExists = 1
                        break

        # Create or open the file:
        if not fileExists:
            fptr = self._createFile(filename)
        else:
            iomode = 1  # read-write
            status, fptr = cfitsio.fits_open_file(filename, iomode)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        # Create the new _TableHdu:
        tableHdu = self._createTable(fptr, keywords, colinfos)

        # Add the new _TableHdu to the Grouping Table:
        if self._groupingTable:
            self._groupingTable.addMemberTableHdu(fptr, keywords)

        # Add the new _TableHdu to the
        self._tableHdus.append(tableHdu)

        # Export the new _TableHdu:
        tableHdu.exprt()

        # Flush the file and reimport to synchronize keywords:
        status = cfitsio.fits_flush_file(fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        tableHdu.imprt()

        return tableHdu

    # -----------------------------------------------------------------

    def _createFile(self, filename, isDatasetFile=0):
        # Create Directory if necessary:
        (dir, file) = os.path.split(filename)
        if not os.path.isdir(dir):
            os.makedirs(dir, mode=0o755)

        # Create Fits file:
        status, fptr = cfitsio.fits_create_file(filename)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Create primary image in file
        status = cfitsio.fits_create_img(fptr, 32, 0, [])
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        status, cpHduNum = cfitsio.fits_movabs_hdu(fptr, 1)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        if not isDatasetFile:
            # Add keyword to primary image:
            keyname = "MBFITS"
            keyval = self.getName().split('/')[-1]
            keycom = "Name of MBFitsFile"
            status = cfitsio.fits_update_key_str(fptr,
                                                 keyname,
                                                 keyval,
                                                 keycom)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        return fptr

    # -----------------------------------------------------------------

    def _createTable(self, fptr, keywords, colinfos):
        # Get the extension name from the keywords:
        extname = ""
        for keyword in keywords:
            if keyword.getName() == "EXTNAME":
                extname = keyword.getValue()
                break

        # From fitsio.h:
        # define IMAGE_HDU  0  /* Primary Array or IMAGE HDU */
        # define ASCII_TBL  1  /* ASCII table HDU  */
        # define BINARY_TBL 2  /* Binary table HDU */
        # define ANY_HDU   -1  /* matches any HDU type */
        binary_tbl = 2
        status = cfitsio.fits_create_tbl(fptr,
                                         binary_tbl,
                                         0,
                                         0,
                                         [],
                                         [],
                                         [],
                                         extname)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        tableHdu = _TableHdu(fptr, self, self._iomode)

        for keyword in keywords:
            tableHdu.addKeyword(keyword)

        for colinfo in colinfos:
            tableHdu.addColumn(colinfo)

        return tableHdu

    # -----------------------------------------------------------------

    def _createGroupingTable(self, filename, groupName):
        status, fptr = cfitsio.fits_open_file(filename, self._iomode)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        groupType = _GroupingHdu.getGroupType()
        status = cfitsio.fits_create_group(fptr, groupName, groupType)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        groupingTable = _GroupingHdu(fptr, self, self._iomode)
        return groupingTable

# ---------------------------------------------------------------------
# ---- Class _Hdu -----------------------------------------------------
# ---------------------------------------------------------------------


class _Hdu:
    # NAM: _Hdu (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, fptr, dataset, iomode=0):
        self._dataset = dataset
        self._fptr = fptr
        self._header = _Header(fptr)

        status, filename = cfitsio.fits_file_name(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        if os.path.isfile(filename):
            self._filename = filename
        else:
            gzFilename = filename + ".gz"
            if os.path.isfile(gzFilename):
                self._filename = gzFilename
            else:
                raise MBFitsError(
                    "Hdu %s, %d: File does not exist" %
                    (filename, self._hdunum))

        dummy, self._hdunum = cfitsio.fits_get_hdu_num(
            self._fptr)  # Strange output from cfitsio routine

        if iomode:
            if not dataset.isWriteOpen():
                raise MBFitsError(
                    "Hdu %s, %d: Dataset not open for write" %
                    (self._filename, self._hdunum))
        self._iomode = iomode

    # -----------------------------------------------------------------

    def isOpen(self):
        if self._fptr:
            return 1
        else:
            return 0

    # -----------------------------------------------------------------

    def isWriteOpen(self):
        return self.isOpen() and self._iomode

    # -----------------------------------------------------------------

    def open(self, iomode=0):
        if self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d already open" %
                (self._filename, self._hdunum))

        if not self._dataset.isOpen():
            raise MBFitsError(
                "Hdu %s, %d: Dataset not open" %
                (self._filename, self._hdunum))

        if iomode:
            if not self._dataset.isWriteOpen():
                raise MBFitsError(
                    "Hdu %s, %d: Dataset not open for write" %
                    (self._filename, self._hdunum))
        self._iomode = iomode

        status, fptr = cfitsio.fits_open_file(self._filename, self._iomode)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        if not fptr:
            raise MBFitsError("Failed to open file %s" % self._filename)

        status, cpHduNum = cfitsio.fits_movabs_hdu(fptr, self._hdunum)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        self._header.open(fptr)

        self._fptr = fptr

    # -----------------------------------------------------------------

    def close(self):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        self._header.close()

        status = cfitsio.fits_close_file(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        self._fptr = None

    # -----------------------------------------------------------------

    def imprt(self):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        self._header.imprt()

    # -----------------------------------------------------------------

    def exprt(self):
        if not self.isWriteOpen():
            raise MBFitsError(
                "Hdu %s, %d not open to write" %
                (self._filename, self._hdunum))

        self._header.exprt()

    # -----------------------------------------------------------------

    def getFilename(self):
        return self._filename

    # -----------------------------------------------------------------

    def getKeywordNames(self):
        return self._header.getKeywordNames()

    # -----------------------------------------------------------------

    def getKeyword(self, key):
        return self._header.getKeyword(key)

    # -----------------------------------------------------------------

    def addKeyword(self, keyword):
        if not self.isWriteOpen():
            raise MBFitsError(
                "Hdu %s, %d not open to write" %
                (self._filename, self._hdunum))

        self._header.addKeyword(keyword)

# ---------------------------------------------------------------------
# ---- Class _TableHdu ------------------------------------------------
# ---------------------------------------------------------------------


class _TableHdu(_Hdu):
    # NAM: _TableHdu (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, fptr, dataset, iomode=0):
        _Hdu.__init__(self, fptr, dataset, iomode)

        self._columns = {}  # key: colname;   value: _Column-object
        self._colnames = {}  # key: colnumber; value: colname

        self._wasExported = 0

        self._selection = None

    # -----------------------------------------------------------------

    def open(self, iomode=0):
        _Hdu.open(self, iomode)

        for column in self._columns.values():
            column.open(self._fptr, iomode)

    # -----------------------------------------------------------------

    def close(self):
        if not self.isOpen():
            raise MBFitsError(
                "Table %s, %d not open" %
                (self._filename, self._hdunum))

        if self.hasSelection():
            self.clearSelection()

        for column in self._columns.values():
            column.close()

        _Hdu.close(self)

    # -----------------------------------------------------------------

    def imprt(self):
        _Hdu.imprt(self)

        status, nCols = cfitsio.fits_get_num_cols(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        for iCol in np.arange(1, nCols + 1):
            colinfo = self._importColinfo(iCol)
            col = _Column(self._fptr, iCol, colinfo, iomode=self._iomode)
            self._columns[colinfo.name] = col
            self._colnames[iCol] = colinfo.name

    # -----------------------------------------------------------------

    def exprt(self):
        _Hdu.exprt(self)

        if not self._wasExported:
            # Create the columns only during the first exprt:
            nCols = len(self._columns)
            for iCol in np.arange(1, nCols + 1):
                self._exportColinfo(iCol)

        self._wasExported += 1

    # -----------------------------------------------------------------

    def _importColinfo(self, colnumber):
        # Get some parameters here:
        status, name, dmyUnit, datatype, repeat, dmyScale, dmyZero, dmyNulval, dmyTdisp = \
            cfitsio.fits_get_bcolparms(self._fptr, colnumber)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Check whether TDIMn keyword exists:
        tdimExists = 1
        status, strTDim, comment = \
            cfitsio.fits_read_keyword(self._fptr, "TDIM" + str(colnumber))
        if status:
            if status == 202:
                # KEY_NO_EXIST
                tdimExists = 0
            else:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        if tdimExists:
            status, tdim = cfitsio.fits_read_tdim(self._fptr, colnumber, 100)
            if status:
                # For variable length columns, status indicates an error,
                # while tdim is correct anyway. Hence, ignore status here:
                pass

            # Numeric and Fits differ in the ordering of indices:
            tdim.reverse()
        else:
            tdim = []

        # Get unit and description from TTYPEn keyword:
        status, dmyName, description = \
            cfitsio.fits_read_keyword(self._fptr, "TTYPE" + str(colnumber))
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        status, unit = \
            cfitsio.fits_read_key_unit(self._fptr, "TTYPE" + str(colnumber))
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        colinfo = ColumnInfo()
        colinfo.name = name
        colinfo.datatype = datatype
        colinfo.repeat = repeat
        colinfo.dim = tdim
        colinfo.unit = unit
        colinfo.description = description

        return colinfo

    # -----------------------------------------------------------------

    def _exportColinfo(self, colnumber):
        col = self._columns[self._colnames[colnumber]]
        colinfo = col.getColinfo()

        # Add column to table:
        ttype = colinfo.name
        tform = colinfo.datatype
        # Modify tform if necessary:
        if colinfo.dim:
            repeat = 1
            for idim in colinfo.dim:
                repeat *= idim
        elif colinfo.repeat:
            repeat = colinfo.repeat
        else:
            repeat = ''
        tform = str(repeat) + tform

        status = cfitsio.fits_insert_col(self._fptr,
                                         colnumber,
                                         ttype,
                                         tform)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Set TDIMn keyword if necessary:
        if colinfo.dim:
            tdim = list(colinfo.dim)
             # Numeric and Fits differ in the ordering of indices:
            tdim.reverse()
            if len(tdim) > 0:
                strtdim = "("
                for i in tdim:
                    strtdim += str(i) + ","
                strtdim = strtdim[:-1] + ")"
            else:
                strtdim = "()"
            status = cfitsio.fits_update_key_str(self._fptr,
                                                 "TDIM" + str(colnumber),
                                                 strtdim,
                                                 "dimension of field")
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        # Modify TTYPEn keyword to hold description and unit:
        if colinfo.description:
            keyname = "TTYPE" + str(colnumber)
            status = cfitsio.fits_update_key_str(self._fptr,
                                                 keyname,
                                                 ttype,
                                                 colinfo.description)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        if colinfo.unit:
            keyname = "TTYPE" + str(colnumber)
            status = cfitsio.fits_write_key_unit(
                self._fptr, keyname, colinfo.unit)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

    # -----------------------------------------------------------------

    def getNumColumns(self):
        return len(self._columns)

    # -----------------------------------------------------------------

    def getColumnNames(self):
        result = []
        # sort according to colnumber:
        for iCol in np.arange(1, len(self._colnames) + 1):
            result.append(self._colnames[iCol])
        return result

    # -----------------------------------------------------------------

    def getColumn(self, colName):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        return self._columns[colName]

    # -----------------------------------------------------------------

    def addColumn(self, colinfo):
        if not self.isWriteOpen():
            raise MBFitsError(
                "Hdu %s, %d not open to write" %
                (self._filename, self._hdunum))

        if colinfo.name in self._columns.keys():
            raise MBFitsError("Keyword %s already present" % colinfo.name)

        iCol = len(self._colnames) + 1
        col = _Column(self._fptr, iCol, colinfo, iomode=self._iomode)
        self._columns[colinfo.name] = col
        self._colnames[iCol] = colinfo.name

    # -----------------------------------------------------------------

    def getNumRows(self):
        if not self.isOpen():
            raise MBFitsError(
                "Table %s, %d not open" %
                (self._filename, self._hdunum))

        status, nRows = cfitsio.fits_get_num_rows(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        return nRows

    # -----------------------------------------------------------------

    def getOptimalRowsize(self):
        if not self.isOpen():
            raise MBFitsError(
                "Table %s, %d not open" %
                (self._filename, self._hdunum))

        status, nrows = cfitsio.fits_get_rowsize(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)
        return nrows

    # -----------------------------------------------------------------

    def setSelection(self, expression, firstRow, numRows):
        if not self.isOpen():
            raise MBFitsError(
                "Table %s, %d not open" %
                (self._filename, self._hdunum))

        if not numRows:
            numRows = self.getNumRows()

        self._selection = _Selection(self._fptr, expression, firstRow, numRows)
        for column in self._columns.values():
            column.setSelection(self._selection)

    # -----------------------------------------------------------------

    def clearSelection(self):
        for column in self._columns.values():
            column.clearSelection()

        self._selection = None

    # -----------------------------------------------------------------

    def hasSelection(self):
        if self._selection:
            return 1
        else:
            return 0

# ---------------------------------------------------------------------
# ---- Class _GroupingHdu ---------------------------------------------
# ---------------------------------------------------------------------


class _GroupingHdu(_Hdu):
    # NAM: _GroupingHdu (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def isGroupingHdu(cls, fptr):
        """
        Check the Keyword EXTNAME:
        """
        status, extname, comment = \
            cfitsio.fits_read_keyword(fptr, 'EXTNAME')
        if status:
            return 0
        return extname == "'GROUPING'"
    isGroupingHdu = classmethod(isGroupingHdu)

    def getGroupType(cls):
        # Grouptype parameters from fitsio.h:
        # define GT_ID_ALL_URI  0
        # define GT_ID_REF      1
        # define GT_ID_POS      2
        # define GT_ID_ALL      3
        # define GT_ID_REF_URI 11
        # define GT_ID_POS_URI 12
        return 12
    getGroupType = classmethod(getGroupType)

    # -----------------------------------------------------------------

    def __init__(self, fptr, dataset, iomode=0):
        _Hdu.__init__(self, fptr, dataset, iomode)

        # Aditional index columns:
        self._indexColTtypes = ["EXTNAME",
                                "SUBSNUM",
                                "FEBE",
                                "BASEBAND"]
        self._indexColTforms = ["30A",
                                "1J",
                                "30A",
                                "1J"]
        self._indexColWriteFunctions = {
            "EXTNAME": "cfitsio.fits_write_col_str",
            "SUBSNUM": "cfitsio.fits_write_col_lng",
            "FEBE": "cfitsio.fits_write_col_str",
            "BASEBAND": "cfitsio.fits_write_col_lng"}
        self._indexColBlankValues = {
            "EXTNAME": "",
            "SUBSNUM": -999,
            "FEBE": "",
            "BASEBAND": -999}

        self._wasExported = 0

    # -----------------------------------------------------------------

    def exprt(self):
        _Hdu.exprt(self)

        if not self._wasExported:
            # Add the additionla index columns only during the first exprt:
            status, nCols = cfitsio.fits_get_num_cols(self._fptr)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            status = cfitsio.fits_insert_cols(self._fptr,
                                              nCols + 1,
                                              len(self._indexColTtypes),
                                              self._indexColTtypes,
                                              self._indexColTforms)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

        self._wasExported += 1

    # -----------------------------------------------------------------

    def importMemberTableHdus(self):
        """
        Impoorts all Tables of the Group
        """
        if not self.isOpen():
            raise MBFitsError(
                "GroupingTable %s, %d not open" %
                (self._filename, self._hdunum))

        memberTableHdus = []

        status, nGroupMembers = cfitsio.fits_get_num_members(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Open the members and get their file pointers at the same time:
        for iMember in np.arange(1, nGroupMembers + 1):
            status, fptrMember = \
                cfitsio.fits_open_member(self._fptr, iMember)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            # Create _TableHdu for this member:
            tableHdu = _TableHdu(fptrMember, self._dataset, self._iomode)
            tableHdu.imprt()
            tableHdu.close()
            memberTableHdus.append(tableHdu)

        return memberTableHdus

    # -----------------------------------------------------------------

    def addMemberTableHdu(self, fptrMember, keywords):
        if not self.isWriteOpen():
            raise MBFitsError(
                "Hdu %s, %d not open to write" %
                (self._filename, self._hdunum))

        dummy, hdunumMember = cfitsio.fits_get_hdu_num(
            fptrMember)  # Strange output from cfitsio routine

        status = cfitsio.fits_add_group_member(
            self._fptr, fptrMember, hdunumMember)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        status, numrows = cfitsio.fits_get_num_rows(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        # Fill indexing columns:
        keywordDict = {}
        for keyword in keywords:
            keywordDict[keyword.getName()] = keyword

        for colName in self._indexColTtypes:
            if colName in keywordDict.keys():
                colValue = keywordDict[colName].getValue()
            else:
                colValue = self._indexColBlankValues[colName]

            # String matching parameters from fitsio.h:
            # define CASESEN   1   /* do case-sensitive string match */
            # define CASEINSEN 0   /* do case-insensitive string match */
            casesen = 1
            status, colnum = cfitsio.fits_get_colnum(
                self._fptr, casesen, colName)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            args = "(self._fptr, colnum, numrows, 1, 1, [colValue])"
            status = eval(self._indexColWriteFunctions[colName] + args)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

# ---------------------------------------------------------------------
# ---- Class _Header --------------------------------------------------
# ---------------------------------------------------------------------


class _Header:
    # NAM: _Header (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, fptr):
        self._fptr = fptr

        self._keywords = {}  # key: keyword name;   value: Keyword-object
        self._keywordNames = {}  # key: keyword number; value: keyword name;
                                # keyword name may be None. hence the length
                                # of the two dictionaries may be different!

    # -----------------------------------------------------------------

    def isOpen(self):
        if self._fptr:
            return 1
        else:
            return 0

    # -----------------------------------------------------------------

    def open(self, fptr):
        self._fptr = fptr

    # -----------------------------------------------------------------

    def close(self):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        self._fptr = None

    # -----------------------------------------------------------------

    def imprt(self):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        status, nKeys, nMoreKeys = cfitsio.fits_get_hdrspace(self._fptr)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        for iKey in np.arange(1, nKeys + 1):
            # This does not work for "HIERARCH ESO" keywords:
            # status, keyname, value, comment =
            # cfitsio.fits_read_keyn(self._fptr, iKey)

            status, record = cfitsio.fits_read_record(self._fptr, iKey)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            if not "=" in record or record[:7] == "COMMENT":
                # COMMENT, no keyword
                self._keywordNames[iKey] = None
                continue

            keyname = record.split("=")[0].strip()

            status, value, comment = cfitsio.fits_read_keyword(
                self._fptr, keyname)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError(
                    "CFITSIO: %s" %
                    msg + "(" + str(keyname) + ")", status)

            if len(value) == 0:
                # special case
                datatype = 'A'
                value = ''
            else:
                # Get keytype, translate keytype to datatype,
                # and convert value to corresponding type:
                status, keytype = cfitsio.fits_get_keytype(value)
                if status:
                    msg = cfitsio.fits_get_errstatus(status)
                    raise MBFitsError("CFITSIO: %s" % msg, status)

                if keytype == 'C':
                    datatype = 'A'
                    value = value[1:-1].strip()
                elif keytype == 'L':
                    datatype = 'L'
                    if value == 'T':
                        value = 1
                    else:
                        value = 0
                elif keytype == 'I':
                    datatype = 'J'
                    value = int(value)
                elif keytype == 'F':
                    datatype = 'D'
                    value = np.float(value)
                else:
                    raise MBFitsError(
                        "Invalid keytype %s for keyword %s" %
                        (keytype, keyname))

            # Get unit:
            status, unit = \
                cfitsio.fits_read_key_unit(self._fptr, keyname)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            # Create keyword:
            kw = Keyword(name=keyname,
                         value=value,
                         comment=comment,
                         datatype=datatype,
                         unit=unit)

            self._keywords[keyname] = kw
            self._keywordNames[iKey] = keyname

    # -----------------------------------------------------------------

    def exprt(self):
        if not self.isOpen():
            raise MBFitsError(
                "Hdu %s, %d not open" %
                (self._filename, self._hdunum))

        nKeys = len(self._keywordNames)
        for iKey in np.arange(1, nKeys + 1):
            keywordName = self._keywordNames[iKey]
            if keywordName:
                kw = self._keywords[keywordName]
                if kw:
                    if 'A' in kw.getDatatype():
                        if kw.getFormat():
                            value = kw.getValue()[:kw.getFormat()]
                        else:
                            value = kw.getValue()
                        status = cfitsio.fits_update_key_str(self._fptr,
                                                             kw.getName(),
                                                             value,
                                                             kw.getComment())
                    elif 'L' in kw.getDatatype():
                        status = cfitsio.fits_update_key_log(self._fptr,
                                                             kw.getName(),
                                                             kw.getValue(),
                                                             kw.getComment())
                    elif 'I' in kw.getDatatype():
                        status = cfitsio.fits_update_key_lng(self._fptr,
                                                             kw.getName(),
                                                             kw.getValue(),
                                                             kw.getComment())
                    elif 'J' in kw.getDatatype():
                        status = cfitsio.fits_update_key_lng(self._fptr,
                                                             kw.getName(),
                                                             kw.getValue(),
                                                             kw.getComment())
                    elif 'E' in kw.getDatatype():
                        if kw.getFormat():
                            format = kw.getFormat()
                        else:
                            format = -7
                        status = cfitsio.fits_update_key_flt(self._fptr,
                                                             kw.getName(),
                                                             kw.getValue(),
                                                             format,
                                                             kw.getComment())
                    elif 'D' in kw.getDatatype():
                        if kw.getFormat():
                            format = kw.getFormat()
                        else:
                            format = -7
                        status = cfitsio.fits_update_key_dbl(self._fptr,
                                                             kw.getName(),
                                                             kw.getValue(),
                                                             format,
                                                             kw.getComment())
                    else:
                        msg = "Invalid datatype %s in keyword %s" % (
                            kw.getDatatype(), kw.getName())
                        raise MBFitsError(msg)
                    if status:
                        msg = cfitsio.fits_get_errstatus(status)
                        raise MBFitsError("CFITSIO: %s" % msg, status)

                    if kw.getUnit():
                        status = cfitsio.fits_write_key_unit(
                            self._fptr, kw.getName(), kw.getUnit())
                        if status:
                            msg = cfitsio.fits_get_errstatus(status)
                            raise MBFitsError("CFITSIO: %s" % msg, status)

    # -----------------------------------------------------------------

    def getKeywordNames(self):
        keynames = []

        iKeys = sorted(self._keywordNames.keys())
        for iKey in iKeys:
            keyname = self._keywordNames[iKey]
            if keyname is not None:
                keynames.append(keyname)
        return keynames

    # -----------------------------------------------------------------

    def getKeyword(self, keyname):
        if keyname in self._keywords.keys():
            return self._keywords[keyname]
        else:
            return None

    # -----------------------------------------------------------------

    def addKeyword(self, keyword):
        if keyword.getName() in self._keywords.keys():
            raise MBFitsError("Keyword %s already present" % keyword.getName())

        iKey = len(self._keywordNames) + 1
        self._keywordNames[iKey] = keyword.getName()
        self._keywords[keyword.getName()] = keyword

# ---------------------------------------------------------------------
# ---- Class _Column --------------------------------------------------
# ---------------------------------------------------------------------


class _Column:
    # NAM: _Column (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, fptr, colnum, colinfo, iomode=0):
        self._fptr = fptr
        self._colnum = colnum
        self._colinfo = colinfo
        self._iomode = iomode

        if 'A' in self._colinfo.datatype:
            self._readFunction = "cfitsio.fits_read_col_str"
            self._writeFunction = "cfitsio.fits_write_col_str"
            self._numericTypecode = None
        elif 'L' in self._colinfo.datatype:
            self._readFunction = "cfitsio.fits_read_col_log"
            self._writeFunction = "cfitsio.fits_write_col_log"
            self._numericTypecode = np.int0
        elif 'B' in self._colinfo.datatype:
            # 'B'
            self._readFunction = "cfitsio.fits_read_col_byt"
            self._writeFunction = "cfitsio.fits_write_col_byt"
            self._numericTypecode = np.uint8
        elif ('I' in self._colinfo.datatype):
            # 'I', 'UI'
            self._readFunction = "cfitsio.fits_read_col_int"
            self._writeFunction = "cfitsio.fits_write_col_int"
            self._numericTypecode = np.int16
        elif ('K' in self._colinfo.datatype):
            # 'K', 'UK'
            self._readFunction = "cfitsio.fits_read_col_lng"
            self._writeFunction = "cfitsio.fits_write_col_lng"
            self._numericTypecode = np.int32
        elif ('J' in self._colinfo.datatype):
            # 'J', 'JJ', 'UJ'
            self._readFunction = "cfitsio.fits_read_col_lng"
            self._writeFunction = "cfitsio.fits_write_col_lng"
            self._numericTypecode = np.int32
        elif ('E' in self._colinfo.datatype):
            self._readFunction = "cfitsio.fits_read_col_flt"
            self._writeFunction = "cfitsio.fits_write_col_flt"
            self._numericTypecode = np.float32
        elif ('F' in self._colinfo.datatype):
            if ('FC' in self._colinfo.datatype):
                self._readFunction = None
                self._writeFunction = None
                self._numericTypecode = None
            elif ('FM' in self._colinfo.datatype):
                self._readFunction = None
                self._writeFunction = None
                self._numericTypecode = None
            else:
                self._readFunction = "cfitsio.fits_read_col_flt"
                self._writeFunction = "cfitsio.fits_write_col_flt"
                self._numericTypecode = np.float32
        elif ('D' in self._colinfo.datatype):
            self._readFunction = "cfitsio.fits_read_col_dbl"
            self._writeFunction = "cfitsio.fits_write_col_dbl"
            self._numericTypecode = np.float64
        elif ('G' in self._colinfo.datatype):
            self._readFunction = "cfitsio.fits_read_col_dbl"
            self._writeFunction = "cfitsio.fits_write_col_dbl"
            self._numericTypecode = np.float64
        else:
            self._readFunction = None
            self._writeFunction = None
            self._numericTypecode = None

        self._selection = None

    # -----------------------------------------------------------------

    def isOpen(self):
        if self._fptr:
            return 1
        else:
            return 0

    # -----------------------------------------------------------------

    def isWriteOpen(self):
        return self.isOpen() and self._iomode

    # -----------------------------------------------------------------

    def open(self, fptr, iomode=0):
        self._fptr = fptr
        self._iomode = iomode

    # -----------------------------------------------------------------

    def close(self):
        if not self.isOpen():
            raise MBFitsError("Column %d not open" % self._colnum)

        self._fptr = None

    # -----------------------------------------------------------------

    def getColnum(self):
        return self._colnum

    # -----------------------------------------------------------------

    def getColinfo(self):
        return self._colinfo

    # -----------------------------------------------------------------

    def getName(self):
        return self._colinfo.name

    # -----------------------------------------------------------------

    def getDatatype(self):
        return self._colinfo.datatype

    # -----------------------------------------------------------------

    def getRepeat(self):
        return self._colinfo.repeat

    # -----------------------------------------------------------------

    def getDim(self):
        return self._colinfo.dim

    # -----------------------------------------------------------------

    def getDescription(self):
        return self._colinfo.description

    # -----------------------------------------------------------------

    def getUnit(self):
        return self._colinfo.unit

    # -----------------------------------------------------------------

    def setSelection(self, selection):
        if not self.isOpen():
            raise MBFitsError("Column %d not open" % self._colnum)

        self._selection = selection

    # -----------------------------------------------------------------

    def clearSelection(self):
        self._selection = None

    # -----------------------------------------------------------------

    def hasSelection(self):
        if self._selection:
            return 1
        else:
            return 0

    # -----------------------------------------------------------------

    def read(self, firstRow, numRows):
        if not self.isOpen():
            raise MBFitsError("Column %s not open" % self._colinfo.name)

        if not self._readFunction:
            return None

        status, numRowsPresent = cfitsio.fits_get_num_rows(self._fptr)

        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        if not numRows:
            numRows = numRowsPresent

        if firstRow + numRows > numRowsPresent + 1:
            numRows = numRowsPresent + 1 - firstRow

        if self._selection:
            selectedRows = self._selection.getSelectedRows()
            if 'A' in self._colinfo.datatype:
                return self._readSelectionA(firstRow, numRows, selectedRows)
            else:
                if 'P' in self._colinfo.datatype:
                    return self._readSelectionP(firstRow, numRows, selectedRows)
                else:
                    return self._readSelection(firstRow, numRows, selectedRows)
        else:
            if 'P' in self._colinfo.datatype:
                if 'A' in self._colinfo.datatype:
                    return self._readPA(firstRow, numRows)
                else:
                    return self._readP(firstRow, numRows)
            else:
                if 'A' in self._colinfo.datatype:
                    return self._readA(firstRow, numRows)
                else:
                    return self._read(firstRow, numRows)

    # -----------------------------------------------------------------

    def _read(self, firstRow, numRows):
        """
        Returns np.array
        """
        if self._colinfo.dim:
            repeat = 1
            for idim in self._colinfo.dim:
                repeat *= idim
        elif self._colinfo.repeat:
            repeat = self._colinfo.repeat
        else:
            repeat = 1

        args = "(self._fptr, self._colnum, firstRow, 1, numRows*repeat, 0)"
        status, result = eval(self._readFunction + args)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        result = self._toNumericArray(result)

        shape = [numRows]
        if self._colinfo.dim:
            shape.extend(self._colinfo.dim)
        else:
            if self._colinfo.repeat > 1:
                shape.extend([self._colinfo.repeat])
        result.shape = shape

        return result

    # -----------------------------------------------------------------

    def _readA(self, firstRow, numRows):
        """
        Returns list of strings
        """
        args = "(self._fptr, self._colnum, firstRow, 1, numRows, 0)"
        status, result = eval(self._readFunction + args)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        if not result:
            result = []

        if not isinstance(result, type([])):
            result = [result]

        return result

    # -----------------------------------------------------------------

    def _readP(self, firstRow, numRows):
        """
        Returns list of np.arrays
        """
        result = []
        for iRow in np.arange(firstRow, firstRow + numRows):
            status, rowRepeat, offset = cfitsio.fits_read_descript(self._fptr,
                                                                   self._colnum,
                                                                   iRow)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            args = "(self._fptr, self._colnum, iRow, 1, rowRepeat, 0)"
            status, rowResult = eval(self._readFunction + args)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            rowResult = self._toNumericArray(rowResult)
            result.append(rowResult)

        return result

    # -----------------------------------------------------------------

    def _readPA(self, firstRow, numRows):
        """
        Returns list of strings.
        """
        result = []
        for iRow in np.arange(firstRow, firstRow + numRows):
            args = "(self._fptr, self._colnum, iRow, 1, 1, 0)"
            status, rowResult = eval(self._readFunction + args)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            result.append(rowResult)

        return result

    # -----------------------------------------------------------------

    def _readSelection(self, firstRow, numRows, selectedRows):
        """
        Returns list of np.arrays
        """
        result = []

        if self._colinfo.dim:
            repeat = 1
            for idim in self._colinfo.dim:
                repeat *= idim
        elif self._colinfo.repeat:
            repeat = self._colinfo.repeat
        else:
            repeat = 1

        for iRow in np.arange(firstRow, firstRow + numRows):
            if iRow in selectedRows:
                args = "(self._fptr, self._colnum, iRow, 1, repeat, 0)"
                status, rowResult = eval(self._readFunction + args)
                if status:
                    msg = cfitsio.fits_get_errstatus(status)
                    raise MBFitsError("CFITSIO: %s" % msg, status)

                rowResult = self._toNumericArray(rowResult)
                result.append(rowResult)

        shape = [len(result)]
        if self._colinfo.dim:
            shape.extend(self._colinfo.dim)
        else:
            if self._colinfo.repeat > 1:
                shape.extend([self._colinfo.repeat])
        result = np.array(result)
        result.shape = shape

        return result

    # -----------------------------------------------------------------

    def _readSelectionA(self, firstRow, numRows, selectedRows):
        """
        Returns list of strings
        """
        result = []
        for iRow in np.arange(firstRow, firstRow + numRows):
            if iRow in selectedRows:
                args = "(self._fptr, self._colnum, iRow, 1, 1, 0)"
                status, rowResult = eval(self._readFunction + args)
                if status:
                    msg = cfitsio.fits_get_errstatus(status)
                    raise MBFitsError("CFITSIO: %s" % msg, status)

                result.append(rowResult)

        return result

    # -----------------------------------------------------------------

    def _readSelectionP(self, firstRow, numRows, selectedRows):
        """
        Returns list of np.arrays
        """
        result = []
        for iRow in np.arange(firstRow, firstRow + numRows):
            if iRow in selectedRows:
                status, rowRepeat, offset = cfitsio.fits_read_descript(
                    self._fptr,
                                                                       self._colnum,
                                                                       iRow)
                if status:
                    msg = cfitsio.fits_get_errstatus(status)
                    raise MBFitsError("CFITSIO: %s" % msg, status)

                args = "(self._fptr, self._colnum, iRow, 1, rowRepeat, 0)"
                status, rowResult = eval(self._readFunction + args)
                if status:
                    msg = cfitsio.fits_get_errstatus(status)
                    raise MBFitsError("CFITSIO: %s" % msg, status)

                rowResult = self._toNumericArray(rowResult)
                result.append(rowResult)

        return result

    # -----------------------------------------------------------------

    def write(self, firstRow, data):
        if not self.isWriteOpen():
            raise MBFitsError(
                "Column %s not open to write" %
                self._colinfo.name)

        if self._writeFunction:
            if 'P' in self._colinfo.datatype:
                if 'A' in self._colinfo.datatype:
                    self._writePA(firstRow, data)
                else:
                    self._writeP(firstRow, self._toNumericArray(data))
            else:
                if 'A' in self._colinfo.datatype:
                    self._writeA(firstRow, data)
                else:
                    self._write(firstRow, self._toNumericArray(data))

    # -----------------------------------------------------------------

    def _write(self, firstRow, data):
        """
        data must be np.array
        """
        # Cast to proper type if necessary:
        if data.dtype.char != self._numericTypecode:
            coldata = data.astype(self._numericTypecode)
        else:
            coldata = data

        args = "(self._fptr, self._colnum, firstRow, 1, np.size(coldata), coldata)"
        status = eval(self._writeFunction + args)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

    # -----------------------------------------------------------------

    def _writeA(self, firstRow, data):
        """
        data must be list of strings
        """
        args = "(self._fptr, self._colnum, firstRow, 1, len(data), data)"
        status = eval(self._writeFunction + args)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

    # -----------------------------------------------------------------

    def _writeP(self, firstRow, data):
        """
        data must be list of np.arrays
        """
        rownum = firstRow
        for dataelement in data:
            # Cast to proper type if necessary:
            if dataelement.dtype.char != self._numericTypecode:
                rowdata = dataelement.astype(self._numericTypecode)
            else:
                rowdata = dataelement

            args = "(self._fptr, self._colnum, rownum, 1, len(rowdata), rowdata)"
            status = eval(self._writeFunction + args)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            rownum += 1

    # -----------------------------------------------------------------

    def _writePA(self, firstRow, data):
        """
        data must be list of strings
        """
        rownum = firstRow
        for rowdata in data:
            args = "(self._fptr, self._colnum, rownum, 1, len(rowdata), rowdata)"
            status = eval(self._writeFunction + args)
            if status:
                msg = cfitsio.fits_get_errstatus(status)
                raise MBFitsError("CFITSIO: %s" % msg, status)

            rownum += 1

    # -----------------------------------------------------------------

    def _toNumericArray(self, x):
        """
        Convert x to a Numeric array, if possible.
        """
        if isinstance(x, type(np.array([]))):
            result = x
        else:
            try:
                # This works for lists and tuples, but raises an exception for
                # scalars
                result = np.array(list(x))
            except:
                # This works for scalars:
                result = np.array([x])
        return result

# ---------------------------------------------------------------------
# ---- Class _Selection -----------------------------------------------
# ---------------------------------------------------------------------


class _Selection:
    # NAM: _Selection (Class)
    # DES: Implementation class; do not use from outside of module ReaMBFits!

    def __init__(self, fptr, expression, firstRow, numRows):
        # self._selectedRows is list of row numbers which conform to the
        # boolean expression
        status, numRowsFound, rowBooleanResults = \
            cfitsio.fits_find_rows(fptr, expression, firstRow, numRows)
        if status:
            msg = cfitsio.fits_get_errstatus(status)
            raise MBFitsError("CFITSIO: %s" % msg, status)

        self._selectedRows = \
            [i + 1 for i in np.arange(len(rowBooleanResults))
             if rowBooleanResults[i] != '0']

    # -----------------------------------------------------------------

    def getSelectedRows(self):
        return self._selectedRows



# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: BogliConfig.py (file)
DES: contains Bogli configuration parameters
"""
__version__ = '$Revision: 2109 $'
__date__ = '$Date: 2007-06-03 14:33:01 +0200 (dim, 03 jun 2007) $'

import os.path
import numpy as np

# lut directory
lutDir = os.path.dirname(__file__) + '/lut/'

# Description of the Label along the X axis
xAxis = {'charheight': 1.0,
         'color': 1,
         'displacement': 3.0,
         'coordinate': 0.5,
         'justification': 0.5,
         'side': 'b',
         'text': ' ',
         'log': 0,
         'draw_it': 1,
         'limits': np.array([0., 0.])}

# Description of the Label along the Y axis, same definition as
# LabelX
yAxis = {'charheight': 1.0,
         'color': 1,
         'displacement': 6.0,
         'coordinate': 0.5,
         'justification': 0.5,
         'side': 'l',
         'text': ' ',
         'log': 0,
         'draw_it': 1,
         'limits': np.array([0., 0.])}

# Description of the Caption, same definition as LabelX
cLabel = {'charheight': 1.0,
          'color': 1,
          'displacement': 1.0,
          'coordinate': 0.5,
          'justification': 0.5,
          'side': 't',
          'text': ' ',
          'draw_it': 1}

# Description of the wedge, same definition as LabelX
zAxis = {'charheight': 2.0,
         'color': 1,
         'displacement': 0.5,
         'side': 'ri',
         'text': ' ',
         'draw_it': 1,
         'width': 2.0,
         'linestyle': 1,
         'linewidth': 1,
         'limits': np.array([0., 0.])}


# Description for the Point style
point = {'symbol': 1,
         'size': 4.0}

# Description for the default line style
line = {'color': 1,
        'linestyle': 1,
        'linewidth': 1}

# Description for the contour lines not used for the moment...
contour = {'charheight': 0.5,
           'color': 1,
           'linestyle': 1,
           'linewidth': 1}

# Description for the boxes
box = {'color': 1,
       'charheight': 1.0,
       'linestyle': 1,
       'linewidth': 1,
       'draw_it': 1}

# Description for the output text
xyouttext = {'charheight': 0.8,
             'color': 5,
             'angle': 0.,
             'justification': 0.5}

# Define the margin around the global viepoint : viewpoint are
# defined in a device-independent manner, using a coordinate
# system whose coordinates run from 0 to 1 in both x and y.

globalViewPoint = {'marging_left': 0.15,
                   'marging_right': 1 - 0.02,
                   'marging_bottom': 0.1,
                   'marging_top': 1 - 0.05,
                   'wedge_size': 0.07}

# marging for clearer plots...
addFracLim = 0.05


def pointSize(size=1):
    """
    DES:    change plotting point size
    INP:    (int) size = point size (default = 1)
    USAGE:  BogliConfig.pointSize(3)
            pointSize(3)
    """
    # 20070603FB created

    point['size'] = size



# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
.. module:: ReaPoint
    :synopsis: contains the Rea Pointing reduction tools
"""
__version__ =  '$Revision: 2719 $'
__date__ =     '$Date: 2010-04-28 15:13:38 +0200 (mer. 28 avril 2010) $'

#----------------------------------------------------------------------------------
#----- Import ---------------------------------------------------------------------
#----------------------------------------------------------------------------------
from rea import ReaMapping
import numpy as np

import os, copy, string

from rea           import ReaConfig
from rea.Bogli     import Plot, Forms, BogliConfig
from rea.Utilities import fitBaseEllipticalGaussian, tolist_rea
from rea.Utilities import compressNan, fitGaussian, modelgaussbase
from rea.ReaError  import ReaError
from rea.fortran   import fUtilities, fMap, fStat

#----------------------------------------------------------------------------------
#----- Rea Pointing Class ----------------------------------------------------------
#----------------------------------------------------------------------------------
class Point(ReaMapping.Map):
    """.. class: Point
    :synopsis: An object of this class is responsible for the reduction of
    pointing scan(s)
    """


    def __init__(self):
        """Initialise an instance."""

        ReaMapping.Map.__init__(self)

        self.Maps = []
        self.__Result = []
        self._PointingResult = 0
        self._ArrayParamOffsets = {}

#--------------------------------------------------------------------------------
#----- public methods -----------------------------------------------------------
#--------------------------------------------------------------------------------

    def iterMap(self, chanList=[], phase=0, flag=0, sizeX=[], sizeY=[]):
        """reconstruct a map in (Az,El) coordinates combining receivers
        and using varying scale to zoom on signal

        Parameters
        ----------
        chanList : list of int
            channels to consider
        phase : int
            phase to plot
        flag  : int
            flag values to consider
        sizeX : float array
            limits in Az of the map
        sizeY : float array
            limits in El of the map
        """

        self.Maps = []  # reinitialise Maps attribute
        # start with one pixel per beam
        self.slowMap(chanList=chanList, phase=phase, flag=flag, sizeX=sizeX,\
                sizeY=sizeY, oversamp=1.)

        firstMap = self._Map__Results
        self.Maps.append(firstMap)

        # localize max. of signal
        dimX = firstMap.shape[0]
        dimY = firstMap.shape[1]

        # convert to list, to use index
        l = tolist_rea(np.ravel(firstMap))

        # remove NaN values, if any
        allNumbers = []
        for x in l:
            if str(x) != str(np.nan):
                allNumbers.append(x)

        maxi = max(allNumbers)
        mini = min(allNumbers)
        # remove the Nan in this list, otherwise index does not work
        for n in range(len(l)):
            if str(l[n]) == str(np.nan):
                l[n] = mini
        posMax = l.index(maxi)

        # convert to x,y
        maxX = int(posMax/dimY)
        maxY = posMax - dimY * maxX + 0.5
        maxX = maxX + 0.5
        # then to Az, El
        maxAz, maxEl = self.wcs2phy(maxX, maxY)
        # new limits for the second map
        newX1, newY1 = self.wcs2phy(maxX-3.5, maxY-3.5)
        newX2, newY2 = self.wcs2phy(maxX+3.5, maxY+3.5)
        Plot.plot([newX1, newX2, newX2, newX1, newX1],\
                  [newY1, newY1, newY2, newY2, newY1], style='l', overplot=1)
        self.MessHand.info(" after 1st iteration, offsets Az,El ="+str(maxAz)+","+str(maxEl))

        # second iteration:
        # compute a second map +/- 3 beams around max
        self.slowMap(chanList=chanList, phase=phase, flag=flag,\
                sizeX=[newX1, newX2], sizeY=[newY1, newY2], oversamp=3.)
        secondMap = self._Map__Results
        self.Maps.append(secondMap)
        # determine new maximum position
        dimX = secondMap.shape[0]
        dimY = secondMap.shape[1]
        # convert to list, to use index
        l = tolist_rea(np.ravel(secondMap))
        # remove NaN values, if any
        allNumbers = []
        for x in l:
            if str(x) != str(np.nan):
                allNumbers.append(x)
        maxi = max(allNumbers)
        mini = min(allNumbers)
        # remove the Nan in this list, otherwise index does not work
        for n in range(len(l)):
            if str(l[n]) == str(np.nan):
                l[n] = mini
        posMax = l.index(maxi)
        # convert to x,y
        maxX = int(posMax/dimY)
        maxY = posMax - dimY * maxX + 0.5
        maxX = maxX + 0.5
        # then to Az, El
        maxAz, maxEl = self.wcs2phy(maxX, maxY)
        # new limits for the third map
        newX1, newY1 = self.wcs2phy(maxX-3.5, maxY-3.5)
        newX2, newY2 = self.wcs2phy(maxX+3.5, maxY+3.5)
        Plot.plot([newX1, newX2, newX2, newX1, newX1],\
                  [newY1, newY1, newY2, newY2, newY1], style='l', overplot=1)
        self.MessHand.info(" after 2nd iteration, offsets Az,El ="+str(maxAz)+","+str(maxEl))

        # third iteration:
        # compute a second map +/- 3 pixels around max
        self.slowMap(chanList=chanList, phase=phase, flag=flag,\
                sizeX=[newX1, newX2], sizeY=[newY1, newY2], oversamp=5.)

        self.Maps.append(self._Map__Results)

    def solvePointing(self, chanList=[], gradient=False, circular=False, radius= -5, \
                      Xpos = 0., Ypos = 0., fixedPos = False, \
                      plot=False, display=True, caption='', style='idl4', aspect=True):
        """compute the offset on timelines

        Parameters
        ----------
        chanList : list of int
            list of channels to be used (default: all)
        gradient : bool
            shall we fit a gradient ? (default: no)
         circular : bool
            fit a cricular gaussian instead of an elliptical gaussian
        radius : float
            use only bolo inside this radius
            (negative means multiple of beam) (default 5 beams)
        Xpos, Ypos : float
            source position if using fixed position
        fixedPos : bool
            if set, don't fit position, but use Xpos, Ypos
        plot : bool
            do we plot the results? (default: no)
        display : bool
            display the result of the fit (default: yes)
        caption : str
            plot caption
        aspect : bool
            do we keep aspect ratio ?

        Notes
        -----

        the results is store in self.PoitingResult (i.e. all
        parameters as computed by mpfit routine).

        If mpfit failed, then self.PoitingResult is set to -1

        """

        # Retrieve the data..
        # The az/el off already contains the actual array offsets
        chanList = self.ReceiverArray.checkChanList(chanList)
        theData  = self.getChanListData('flux',  chanList, dataFlag=[], getFlaggedData=0)
        theWeig  = self.getChanListData('weight', chanList, dataFlag=[], getFlaggedData=0)
        offsets  = self.getChanListData('azeloff', chanList, dataFlag=[], getFlaggedData=0)

        fwhm = self.ReceiverArray.BeamSize

        # Transform that to a flat 1D array
        dataX = np.array(offsets)[:,:, 0].flatten()
        dataY = np.array(offsets)[:,:, 1].flatten()
        theData = np.array(theData).flatten()
        theWeig = np.array(theWeig).flatten()

        if radius < 0:
            radius = abs(radius)*fwhm

        if radius != 0:

            good_for_fit = np.less(np.sqrt((dataX-Xpos)**2+(dataY-Ypos)**2), radius)

            dataX   = np.compress(np.equal(good_for_fit, 1), dataX)
            dataY   = np.compress(np.equal(good_for_fit, 1), dataY)
            theData = np.compress(np.equal(good_for_fit, 1), theData)
            theWeig = np.compress(np.equal(good_for_fit, 1), theWeig)

            # fUtilities.compress crash for array > 2100000
            # dataX,n   = fUtilities.compress(dataX,good_for_fit,1)
            # dataX     = dataX[:n]
            # dataY,n   = fUtilities.compress(dataY,good_for_fit,1)
            # dataY     = dataY[:n]
            # theData,n = fUtilities.compress(theData,good_for_fit,1)
            # theData   = theData[:n]


        # now use the data to fit a 2D-Gaussian

        try:
            self._PointingResult = fitBaseEllipticalGaussian(theData, dataX, dataY,\
                                                            err = 1./np.sqrt(theWeig),\
                                                            fwhm=fwhm,\
                                                            gradient=gradient,\
                                                            circular=circular,\
                                                            Xpos=Xpos, Ypos=Ypos,\
                                                            fixedPos=fixedPos)
            self.showPointing(plot=plot, display=display,
                              caption=caption, aspect=aspect,
                              style=style)
        except ReaError as error:
            self.MessHand.warning('fit did not converge:' + error.msg)
            self._PointingResult = None
        except ValueError:
            self.MessHand.warning('fit was not possible')
            self._PointingResult = None

    def solvePointingOnMap(self, gradient=False, circular=False, radius=-10, \
                           Xpos = 0., Ypos = 0., fixedPos = False, \
                           plot=False, display=True, caption='', aspect=True, style='idl4'):
        """compute the offset on the data.Map object

        Parameters
        ----------
        gradient : bool
            shall we fit a gradient ? (default: no)
         circular : bool
            fit a cricular gaussian instead of an elliptical gaussian
        radius : float
            use only bolo inside this radius
            (negative means multiple of beam) (default 5 beams)
        Xpos, Ypos : float
            source position if using fixed position
        fixedPos : bool
            if set, don't fit position, but use Xpos, Ypos
        plot : bool
            do we plot the results? (default: no)
        display : bool
            display the result of the fit (default: yes)
        caption : str
            plot caption
        aspect : bool
            do we keep aspect ratio ?
        style : str
            the style used for the color

        Notes
        -----
        * The result of the fits is store in self.PointingResult,
        i.e. all parameters as computed by mpfit routine.

        * If mpfit failed, then self.PoitingResult is set to -1

        WARNING : No Smoothing should be applied to the map
        before using this function, or the fitted fwhm will be
        useless, use fine oversamp to make reasonable fit

        """

        if self.Map.Data == []:
            self.MessHand.error('No map computed yet')
            return

        # fit a 2D-Gaussian on the map
        try:
            self._PointingResult = self.Map.extractSource(gradient=gradient,
                                                         circular=circular,
                                                         radius=radius,
                                                         Xpos=Xpos, Ypos=Ypos,
                                                         fixedPos=fixedPos)
            self.showPointing(plot=plot, display=display, caption=caption,
                              aspect=aspect, style=style)
        except ReaError as error:
            self.MessHand.warning('fit did not converge:' + error.msg)
            self._PointingResult = None


    def showPointing(self, plot=True, display=True, noMap=False, caption='',
                     aspect=True, style='idl4', limitsZ=[], noerase=False):
        """display results of last solvePointing (in text, and on the map if plot=1)

        Parameters
        ----------
        plot : bool
            display the results on a map
        display : bool
            display the result on screen
        noMap : bool
            do not display the map
        caption : str
            caption of the plot
        aspect : bool
            do we keep aspect ratio ?
        style : str
            the style used for the color
        limitsZ : float array
            limits in amplitude for the map
        noerase : bool
            do we keep the plot ?
        """

        if not(self._PointingResult):
            self.MessHand.error('No pointing results to be displayed')
            return

        PointingResult = self._PointingResult

        if display:
            # Display out the result

            # Dictionnary to make links between keys in the pointing result
            # variable and the displayed name
            pointingDict = {'gauss_x_offset': 'Delta Az ["]',\
                            'gauss_x_fwhm'  : 'FWHM_1   ["]',\
                            'gauss_y_offset': 'Delta El ["]',\
                            'gauss_y_fwhm'  : 'FWHM_2   ["]',\
                            'gauss_tilt'    : 'Tilt   [deg]',
                            'gauss_peak'    : 'Peak flux   '}

            if self.Map.WCS['CTYPE1'].find('RA') >= 0 and self.Map.WCS['CTYPE2'].find('DEC') >= 0:
                pointingDict['gauss_x_offset'] = 'RA     [deg]'
                pointingDict['gauss_y_offset'] = 'Dec    [deg]'
                # if map is equatorial, convert FWHM to arcsec for display
                PointingResult['gauss_x_fwhm']['value'] *= 3600.
                PointingResult['gauss_y_fwhm']['value'] *= 3600.
            if self.Map.WCS['CTYPE1'].find('GLON') >= 0 and self.Map.WCS['CTYPE2'].find('GLAT') >= 0:
                pointingDict['gauss_x_offset'] = 'GLon   [deg]'
                pointingDict['gauss_y_offset'] = 'GLat   [deg]'
                # if map is in Galactic, convert FWHM to arcsec for display
                PointingResult['gauss_x_fwhm']['value'] *= 3600.
                PointingResult['gauss_y_fwhm']['value'] *= 3600.

            # Select what will be displayed from the PointingResult
            toBeOutput = ['gauss_peak',\
                          'gauss_x_offset', 'gauss_y_offset',\
                          'gauss_x_fwhm', 'gauss_y_fwhm', 'gauss_tilt']

            outStr = ""
            for item in toBeOutput:
                outStr += pointingDict[item] + " = " + \
                         str(PointingResult[item]['value'])+'\n'
            self.MessHand.info(outStr)
            if self.Map.WCS['CTYPE1'].find('RA') >= 0 and self.Map.WCS['CTYPE2'].find('DEC') >= 0:
                # convert FWHM back to deg to compute the ellipsis
                PointingResult['gauss_x_fwhm']['value'] *= 1./3600.
                PointingResult['gauss_y_fwhm']['value'] *= 1./3600.
            if self.Map.WCS['CTYPE1'].find('GLON') >= 0 and self.Map.WCS['CTYPE2'].find('GLAT') >= 0:
                PointingResult['gauss_x_fwhm']['value'] *= 1./3600.
                PointingResult['gauss_y_fwhm']['value'] *= 1./3600.

        if plot:
            # extract parameters of the ellipsis to plot
            xoff = PointingResult['gauss_x_offset']['value']
            yoff = PointingResult['gauss_y_offset']['value']
            xfwhm = PointingResult['gauss_x_fwhm']['value']
            yfwhm = PointingResult['gauss_y_fwhm']['value']
            tilt = PointingResult['gauss_tilt']['value']*np.pi/180.

            if not self.Map:
                self.doMap(style=style)
            else:
                if not noMap:
                    self.showMap(aspect=aspect, style=style, limitsZ=limitsZ,
                                 noerase=noerase, caption=caption)
            # Overlay the pointing result
            Forms.ellipse(xoff, yoff, xfwhm, yfwhm, tilt)


    def arrayParameters(self, chanList='all', gradient=0, circular=0, radius=0, plot=False, onMap=False):
        """determine the array parameters from the data

        Parameters
        ----------
        chanList : list of int
            the channel list to be used (default: current list)
        gradient : bool
            remove a background gradient in the data (default: no)
        circular : bool
            fit a cricular gaussian instead of an elliptical gaussian
        radius : float
            limit the fit to a certain range
        plot : bool
            do we plot the result ?
        onMap : bool
            do we fit maps ? (False, would make the fit on timelines)

        Notes
        -----

        * The results of the fits are stored as an associative array in self._ArrayParamOffsets,
        ie. all parameters as computed by mpfit for all channels

        * If mpfit failed then the corresponding value is set to -1


        """

        chanList = self.ReceiverArray.checkChanList(chanList)

        fwhm = self.ReceiverArray.BeamSize

        arrayParamOffsets = self._ArrayParamOffsets

        # Loop through all the channels
        for chan in chanList:
            self.MessHand.info('Fitting channel %i...'%chan)

            if plot:
                if radius:
                    sizeX = [-1.*radius, 1.*radius]
                    sizeY = sizeX
                else:
                    sizeX = []
                    sizeY = []

                self.doMap([chan], oversamp=5,\
                           sizeX=sizeX, sizeY=sizeY, \
                           aspect=1, noPlot=1, showRms=0)

            # By construction the source should be at 0,0
            if not onMap:
                self.solvePointing([chan], gradient=gradient, circular=circular, radius=radius,\
                                       plot=plot, display=False)
            else:
                self.doMap([chan], oversamp=5, noPlot=1)
                self.solvePointingOnMap(gradient=gradient, circular=circular, radius=radius,\
                                            plot=plot, display=False)


            if plot:
                if self._PointingResult:
                    diffOffX = self._PointingResult['gauss_x_offset']['value']
                    diffOffY = self._PointingResult['gauss_y_offset']['value']
                else:
                    diffOffX = 0
                    diffOffY = 0

                Plot.xyout(diffOffX, diffOffY, 'Channel '+str(chan))
                Plot.nextpage()


            if self._PointingResult:
                self.MessHand.longinfo('Offsets to RCP (%5.1f,%5.1f)"; ' % \
                                       (self._PointingResult['gauss_x_offset']['value'],\
                                        self._PointingResult['gauss_y_offset']['value']) + \
                                       'retrieved FHWM %4.1f"x%4.1f" (%5.2f)' %\
                                        (self._PointingResult['gauss_x_fwhm']['value'],\
                                         self._PointingResult['gauss_y_fwhm']['value'],\
                                         self._PointingResult['gauss_tilt']['value']))

            arrayParamOffsets[chan] = self._PointingResult

        # Normalizing gains by keeping only the good channels, change
        # fixed value to value propto beam size

        self.MessHand.longinfo('Normalizing the gains')
        fluxes = []
        for chan in arrayParamOffsets.keys():
            if arrayParamOffsets[chan] and \
                    arrayParamOffsets[chan]['gauss_x_offset']['value'] < 10 and \
                    arrayParamOffsets[chan]['gauss_y_offset']['value'] < 10 and \
                    arrayParamOffsets[chan]['gauss_x_fwhm']['value'] < 25 and \
                    arrayParamOffsets[chan]['gauss_y_fwhm']['value'] < 25:
                fluxes.append(arrayParamOffsets[chan]['gauss_peak']['value'])

        fluxes     = np.array(fluxes)
        medianFlux = fStat.f_median(fluxes)

        for chan in arrayParamOffsets.keys():
            if arrayParamOffsets[chan]:
                arrayParamOffsets[chan]['gain'] = arrayParamOffsets[chan]['gauss_peak']['value']/medianFlux

        self._ArrayParamOffsets = arrayParamOffsets
        self.MessHand.longinfo('You can now use the ReaPointing.updateArrayParameters method')


    def updateArrayParameters(self):
        """Update the Parameters Offsets with the computed values

        Notes
        -----
        leave the reference chan position unchanged
        """

        arrayParamOffsets = self._ArrayParamOffsets
        offsets = copy.copy(self.ReceiverArray.Offsets)
        refChan = self.ReceiverArray.RefChannel
        gains   = self.ReceiverArray.Gain
        fwhm    = self.ReceiverArray.FWHM
        tilt    = self.ReceiverArray.Tilt

        if fwhm.size == 0:
            nc = self.ReceiverArray.NChannels
            tilt = np.zeros(nc, np.float32)
            fwhm = np.zeros((2, nc), np.float32)

        # find the reference offset
        if refChan in arrayParamOffsets.keys():
            refOffset = np.array([arrayParamOffsets[refChan]['gauss_x_offset']['value'],\
                               arrayParamOffsets[refChan]['gauss_y_offset']['value']])
        else:
            self.MessHand.warning('Reference Offsets not found in fit')
            refOffset = np.array([0, 0])

        for chan in arrayParamOffsets.keys():
            chanIndex = self.ReceiverArray.getChanIndex(chan)
            if not chanIndex in [-1] and arrayParamOffsets[chan]:
                thisOffset = np.array([arrayParamOffsets[chan]['gauss_x_offset']['value'],\
                                    arrayParamOffsets[chan]['gauss_y_offset']['value']])

                offsets[:, chanIndex] = ( offsets[:, chanIndex] - \
                                             np.array([(thisOffset-refOffset)]).transpose().astype(np.float32))

                gains[chanIndex] = arrayParamOffsets[chan]['gain']
                fwhm[:, chanIndex] = np.array([[arrayParamOffsets[chan]['gauss_x_fwhm']['value'], \
                                                arrayParamOffsets[chan]['gauss_y_fwhm']['value']]]).transpose().astype(np.float32)
                tilt[chanIndex]   = arrayParamOffsets[chan]['gauss_tilt']['value']
            else:
                self.MessHand.warning('Channel %s not valid or no good fit'%str(chan))

        self.ReceiverArray.Offsets = offsets
        self.ReceiverArray.Gain    = gains
        self.ReceiverArray.FWHM    = fwhm
        self.ReceiverArray.Tilt    = tilt

    #---------------------------------------------------------------------------------
    def reduce(self, datasetName='', obstoProc=[], febe='', baseband=1,
               radius=-2., update=False, tau=0.):
        """Process a Pointing scan - this method is called by the apexCalibrator

        Parameters
        ----------
        datasetName : str
            path to the dataset to be reduced
        obstoProc : list of int
            list of subscans to consider (default: all)
        febe : str
            frontend-backend to consider
        baseband : int
            which baseband do we reduce ?
        radius : float
            radius to be used for fitting (def: 2xbeam)
        update : bool
            continue previous scan? (def: no)
        tau : float
            zenithal opacity to apply
        """
        if len(obstoProc) == 1:
            if isinstance(obstoProc[0], type([])): # e.g. obstoProc == [range(4,8)]
                self.read(inFile=datasetName, subscans=obstoProc[0],
                          febe=febe, readAzEl0=1, baseband=baseband)
            else:
                # cannot work subscan by subscan
                self.read(inFile=datasetName, subscans=range(1, obstoProc[0]+1),
                          febe=febe, readAzEl0=1, baseband=baseband)
        else:
            self.read(inFile=datasetName, subscans=obstoProc,
                      febe=febe, readAzEl0=1, baseband=baseband)

        # If chopped data, then compute phase diff
        if self.ScanParam.WobUsed:
            self.ScanParam.computeOnOff()
            self._phaseDiff()

        if string.upper(self.ScanParam.Object) != "JUPITER" and \
               string.upper(self.ScanParam.Object) != "MARS" and \
               string.upper(self.ScanParam.Object) != "SATURN":
            self.flagFractionRms(ratio=10)
        self.medianBaseline()
        self.flatfield()

        # TODO: This is WRONG
        # Conversion to Jy - ToDo: not so clean!
        if string.find(self.ReceiverArray.FeBe, 'LABOCA') >= 0:
            be = self.ReceiverArray.BEGain
            self.Data *= np.array(be/270. * 6.3E6, np.float)
        elif string.find(self.ReceiverArray.FeBe, 'BOLOSZ') >= 0:
            self.Data *= np.array(0.135, np.float)
        else:
            self.MessHand.warning("Unknown instrument - data not calibrated")

        if tau:
            self.correctOpacity(tau)

        Plot.panels(2, 2)
        Plot.nextpage()  # start a new page
        # First plot = signal before skynoise removal
        four = self.ReceiverArray.fourpixels()
        self.signal(four, noerase=1, caption=self.ScanParam.caption()+' - Raw signal')

        # Now clean up the signal
        beam = self.ReceiverArray.BeamSize
        if string.upper(self.ScanParam.Object) == "JUPITER":
            self.flagPosition(radius=3.*beam, flag=8)
            self.flagFractionRms(ratio=10)
        else:
            self.flagPosition(radius=2.*beam, flag=8)
            self.flagFractionRms(ratio=5)
        self.medianNoiseRemoval(chanRef=-1, factor=0.95, nbloop=5)
        # Repeat median noise splitting channels in 4 groups
        # (corresponds to amplifier boxes for LABOCA)
        nb_all = self.ReceiverArray.NChannels
        if nb_all > 12:  # at least 3 bolo per group
            self.medianNoiseRemoval(range(int(nb_all/4)+1),
                                    chanRef=-1, factor=0.98, nbloop=2)
            self.medianNoiseRemoval(range(int(nb_all/4)+1, 2*int(nb_all/4)+1),
                                    chanRef=-1, factor=0.98, nbloop=2)
            self.medianNoiseRemoval(range(2*int(nb_all/4)+1, 3*int(nb_all/4)+1),
                                    chanRef=-1, factor=0.98, nbloop=2)
            self.medianNoiseRemoval(range(3*int(nb_all/4)+1, nb_all+1),
                                    chanRef=-1, factor=0.98, nbloop=2)
        self.polynomialBaseline(order=1, subscan=0)
        if string.upper(self.ScanParam.Object) == "JUPITER":
            self.flagFractionRms(ratio=10)
        else:
            self.flagFractionRms(ratio=5)

        self.despike(below=-3, above=5)
        self.computeWeight()
        self.unflag(flag=8)
        Plot.nextpage()
        four = self.ReceiverArray.fourpixels()
        self.signal(four, noerase=1, caption='Skynoise subtracted')

        # Now compute the map and solve for pointing
        self.doMap(oversamp=3., noPlot=1,
                   sizeX=[-10.*beam, 10.*beam], sizeY=[-10.*beam, 10.*beam])
        self.Map.computeRms()
        self.solvePointingOnMap(radius=radius, plot=0, display=0)
        Plot.nextpage()
        # compute good Z limits
        pixels = np.ravel(self.Map.Data)
        pixels, nb = compressNan([pixels])
        minmax = fStat.minmax(pixels)
        minZ = max([minmax[0], -3.*self.Map.RmsBeam])
        if self._PointingResult :
            peak = self._PointingResult['gauss_peak']['value']
            if peak > 3.*self.Map.RmsBeam:
                maxZ = peak
            else:
                maxZ = minmax[1]
        else:
            maxZ = minmax[1]
        if self._PointingResult :
            self.showPointing(noerase=1, limitsZ=[minZ, maxZ], caption=' ')
            offX = self._PointingResult['gauss_x_offset']['value']
            offY = self._PointingResult['gauss_y_offset']['value']
            fwX  = self._PointingResult['gauss_x_fwhm']['value']
            fwY  = self._PointingResult['gauss_y_fwhm']['value']
            Plot.xyout(0, 8.5*beam, str("peak = %6.1f Jy/beam"%(peak)), size=2.)
            Plot.xyout(0, -8.*beam, str("FWHM = %4.1f x %4.1f"%(fwX, fwY)), size=2.)
            Plot.xyout(0, -9.5*beam, str("x=%5.1f ; y=%5.1f"%(offX, offY)), size=2.)
        else:
            # when pointing failed, display only the map
            self.Map.display(noerase=1, limitsZ=[minZ, maxZ], caption=' ')
            Plot.xyout(0, -9.5*beam, "no fit", size=2.5)
        Plot.nextpage()

        # Finally show the RMS of all bolos, and compute median NEFD
        # but flag the source first
        if self._PointingResult != -1:
            if peak > 3.*self.Map.RmsBeam:
                self.flagPosition(Az = offX, El = offY, radius = max([fwX, fwY]))
        self.plotBoloRms(noerase=1)
        rms = self.getChanListData('rms')
        medrms = fStat.f_median(rms)
        sampl = fStat.f_median(1./self.ScanParam.get('deltat'))
        nefd = medrms / np.sqrt(sampl)
        Plot.xyout(0, 1.05*max(self.ReceiverArray.Offsets[1,:]),
                   str("NEFD = %5.1f mJy/sqrt(Hz)"%(nefd*1.E3)), size=1.5)
        Plot.panels(1, 1)

    #---------------------------------------------------------------------------------
    def reduceCross(self,datasetName='',obstoProc=[],febe='',baseband=1,update=False):
        """Process a Pointing scan observed with cross-OTF

        Parameters
        ----------
        datasetName : str
            path to the dataset to be reduced
        obstoProc : list of int
            list of subscans to consider (default: all)
        febe : str
            frontend-backend to consider
        baseband : int
            which baseband do we reduce ?
        update : bool
            continue previous scan? (def: no)
        """
        if len(obstoProc) == 1:
            if isinstance(obstoProc[0], type([])): # e.g. obstoProc == [range(4,8)]
                self.read(inFile=datasetName, subscans=obstoProc[0],
                          febe=febe, readAzEl0=1, baseband=baseband)
            else:
                # cannot work subscan by subscan
                self.read(inFile=datasetName, subscans=range(1, obstoProc[0]+1),
                          febe=febe, readAzEl0=1, baseband=baseband)
        else:
            self.read(inFile=datasetName, subscans=obstoProc,
                      febe=febe, readAzEl0=1, baseband=baseband)

        # First, subtract zero order baseline and average noise
        self.medianBaseline(subscan=0)

        # Plot signal before subtracting correlated noise
        Plot.panels(2, 2)
        Plot.nextpage()  # start a new page
        four = self.ReceiverArray.fourpixels()
        self.signal(four, noerase=1, caption=self.ScanParam.caption()+' - Raw signal')

        # If chopped data, then compute phase diff
        if self.ScanParam.WobUsed:
            self.ScanParam.computeOnOff()
            self._phaseDiff()

        self.medianBaseline(subscan=0)
        ref = self.ReceiverArray.RefChannel
        self.averageNoiseRemoval(chanRef=ref)

        self.despike(below=-5)  # flag negative, if any
        Plot.nextpage()  # start a new page
        self.signal(four, noerase=1, caption='Phase diffed, average noise subtracted')

        # We will store Az and El offsets based on subscans
        offAz, fluxAz = np.array([], np.float), np.array([], np.float)
        offEl, fluxEl = np.array([], np.float), np.array([], np.float)
        for i in range(len(self.ScanParam.SubscanNum)):
            subNum = self.ScanParam.SubscanNum[i]
            tmpAz = self.getChanData('azoff', ref, subscans=[subNum])
            tmpEl = self.getChanData('eloff', ref, subscans=[subNum])
            tmpFlux = self.getChanData('flux', ref, subscans=[subNum])
            # Update Az or El data, depending on SCANDIR
            if string.find(self.ScanParam.ScanDir[i], 'LON') > -1:
                self.MessHand.info("Subscan %i in Az direction"%(subNum))
                offAz = np.concatenate((offAz, tmpAz))
                fluxAz = np.concatenate((fluxAz, tmpFlux))
            elif string.find(self.ScanParam.ScanDir[i], 'LAT') > -1:
                self.MessHand.info("Subscan %i in El direction"%(subNum))
                offEl = np.concatenate((offEl, tmpEl))
                fluxEl = np.concatenate((fluxEl, tmpFlux))
            else:
                self.MessHand.warning("Subscan %i in undefined direction, skipping..."%(subNum))

        fwhm2sigma = 1./(2*np.sqrt(2*np.log(2)))

        if len(offAz):
            err = ones(offAz.shape)
            Plot.nextpage()
            Plot.pointSize(5)
            Plot.plot(offAz, fluxAz, noerase=1, style='l', width=1,
                      labelX='Az offset [arcsec]', labelY='Flux [arb. u.]')
            Plot.plot(offAz, fluxAz, overplot=1, style='p')
            m = fitGaussian(offAz, fluxAz, err, const=1)
            if m.status >= 0:
                resAz = m.params[1]
                ampAz = m.params[0]
                widAz = m.params[2] / fwhm2sigma
                xx = min(offAz) + arange(101)*(max(offAz)-min(offAz))/100.
                yy = modelgaussbase(m.params, xx)
                Plot.plot(xx, yy, overplot=1, ci=2, style='l', width=3)
                Plot.plot([resAz, resAz], [0, m.params[0]], overplot=1, style='l', width=5, ci=3)
                Plot.xyout(resAz, 0., "%5.1f"%(resAz), size=2.5)
                y0 = min(fluxAz) + 0.9*(max(fluxAz)-min(fluxAz))
                Plot.xyout(xx[5], y0, "Az", size=4)
            else:
                resAz, ampAz, widAz = -999, 0, 0
        else:
            resAz, ampAz, widAz = -999, 0, 0

        if len(offEl):
            err = ones(offEl.shape)
            Plot.nextpage()
            Plot.pointSize(5)
            Plot.plot(offEl, fluxEl, noerase=1, style='l', width=1,
                      labelX='El offset [arcsec]', labelY='Flux [arb. u.]')
            Plot.plot(offEl, fluxEl, overplot=1, style='p')
            m = fitGaussian(offEl, fluxEl, err, const=1)
            if m.status >= 0:
                resEl = m.params[1]
                ampEl = m.params[0]
                widEl = m.params[2] / fwhm2sigma
                xx = min(offEl) + arange(101)*(max(offEl)-min(offEl))/100.
                yy = modelgaussbase(m.params, xx)
                Plot.plot(xx, yy, overplot=1, ci=2, style='l', width=3)
                Plot.plot([resEl, resEl], [0, m.params[0]], overplot=1, style='l', width=5, ci=3)
                Plot.xyout(resEl, 0., "%5.1f"%(resEl), size=2.5)
                y0 = min(fluxEl) + 0.9*(max(fluxEl)-min(fluxEl))
                Plot.xyout(xx[95], y0, "El", size=4)
            else:
                resEl, ampEl, widEl = -999, 0, 0
        else:
            resEl, ampEl, widEl = -999, 0, 0

        Plot.panels(1, 1)

        # Store results in self.PointingResult
        result = {'gauss_x_offset':resAz,'gauss_y_offset':resEl,
                  'width_x':widAz,'width_y':widEl,
                  'ampl_x':ampAz,'ampl_y':ampEl}
        self.PointingResult = result


    #---------------------------------------------------------------------------------

    #---------------------------------------------------------------------------------
    def writeModelData(self, filename='LABOCA-ABBA.dat'):
        """generate one line to be written in the .dat file used for
        determining pointing model.

        Parameters
        ----------
        filename : str
            the filename to update

        Notes
        -----
        LABOCA Specific : This will work only if data have been read
        in with readAzEl0 = 1 and after the pointing has been reduced

        """

        param = self.ScanParam
        # we need to find the 1st timestamp that is not flagged
        i0 = 0
        while param.FlagHandler.isSetOnIndex(i0):
            i0 += 1
        sourceAz = param.AntAz0 - param.AzOff[i0]
        sourceEl = param.AntEl0 - param.ElOff[i0]
        encodAz  = param.EncAz0 - param.AzOff[i0]
        encodEl  = param.EncEl0 - param.ElOff[i0]
        PCA      = param.PDeltaCA
        PIE      = param.PDeltaIE
        FCA      = param.FDeltaCA
        FIE      = param.FDeltaIE
        scan     = param.ScanNum
        cCA      = self._PointingResult ['gauss_x_offset']['value'] / 3600.
        cIE      = self._PointingResult ['gauss_y_offset']['value'] / 3600.
        datLine = '%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %i\n' % \
                  (sourceAz, sourceEl, encodAz, encodEl,
                   cCA, cIE, PCA, PIE, FCA, FIE, scan)
        output = file(filename, 'a')
        output.write(datLine)
        output.close



# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: ReaFocus.py (module)
DES: contains the Rea focus class
"""
__version__ = '$Revision: 2593 $'
__date__ = '$Date: 2008-06-27 18:37:59 +0200 (ven, 27 jun 2008) $'

# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------

import numpy as np

from rea import ReaPointing
from rea.Bogli import Plot, MultiPlot, BogliConfig
from rea.Utilities import fitParabola, modelparabola
from rea.fortran import fStat

# ---------------------------------------------------------------------
# ---- Rea Focus Class ------------------------------------------------
# ---------------------------------------------------------------------


class Focus(ReaPointing.Point):

    """
    NAM: Focus (class)
    DES: An object of this class is responsible for the focus reduction
         of single or multiple scans and provides the offsets.
    """

    def __init__(self):
        """
        DES: Initialise an instance
        """
        ReaPointing.Point.__init__(self)
        self.FocusOffset = 0.
        self.FocusError = 0.

    # -------------------------------------------------------------------
    # ---- public methods -----------------------------------------------
    # -------------------------------------------------------------------
    def solveFocus(self, noerase=0, caption=''):
        """
        DES: compute the optimal focus position
        """

        refChan = self.ReceiverArray.RefChannel
        modType = self.ScanParam.ScanType
        nSubscan = len(self.ScanParam.SubscanNum)  # subscans read in

        iflux = np.zeros((nSubscan), np.float32)
        isdev = np.zeros((nSubscan), np.float32)
        ifocus = np.zeros((nSubscan), np.float32)

        for i in range(nSubscan):
            flux = self.getChanData(
                'flux',
                refChan,
                subscans=[self.ScanParam.SubscanNum[i]])
            focus = self.getChanData(
                modType,
                refChan,
                subscans=[self.ScanParam.SubscanNum[i]])
            iflux[i] = fStat.f_median(flux)
            isdev[i] = fStat.f_rms(flux, iflux[i])
            ifocus[i] = fStat.f_median(focus)

        if string.find(modType.upper(), 'TILT') > -1:
            ifocus = ifocus * 3600.
            xLabel = modType + ' Offsets [arcsec]'
        else:
            xLabel = modType + ' Offsets [mm]'
        yLabel = 'flux density [arb u.] Channel :' + str(refChan)

        result = fitParabola(ifocus, iflux, isdev)

        x = arange(101.) / 100. * (max(ifocus) - min(ifocus)) + min(ifocus)
        y = modelparabola(result.params, x)

        if not caption:
            caption = self.ScanParam.caption()
        Plot.plot(ifocus, iflux, style='p',
                  labelX=xLabel, labelY=yLabel,
                  caption=caption, noerase=noerase)

        # TODO : need plot with error bar.
        # Plot.plot(ifocus,iflux,style='p',ci=1,overplot=1)
        Plot.plot(x, y, overplot=1, style='l')
        if result.params[2]:
            self.FocusOffset = result.params[1] / (-2.0 * result.params[2])
            if result.perror:
                self.FocusError = result.perror[1] / (-2.0 * result.params[2])
            else:
                self.FocusError = 0.
            self.MessHand.info("Offset : %5.2f +- %5.2f"
                               % (self.FocusOffset, self.FocusError))
            self.MessHand.longinfo("Parabola: a + b x + c x^2, " +
                                   "a = %7.4f b = %7.4f c = %7.4f"
                                   % (result.params[0], result.params[1], result.params[2]))
        else:
            self.MessHand.warning("no focus offset found")
            self.FocusOffset = 0.
            self.FocusError = 0.

    def reduce(self, datasetName='', obstoProc=[], febe='', baseband=1):
        """
        DES: Process a Focus scan - this method is called by the apexCalibrator
        INP: (string) datasetName: path to the dataset to be reduced
             (i list)   obstoProc: list of subscans to consider (default: all)
        """
        if len(obstoProc) == 1:
            if isinstance(obstoProc[0], type([])):  # e.g. obstoProc == [range(4,8)]
                self.read(inFile=datasetName, subscans=obstoProc[0],
                          febe=febe, baseband=baseband)
            else:
                # cannot work subscan by subscan
                self.read(
                    inFile=datasetName, subscans=range(1, obstoProc[0] + 1),
                          febe=febe, baseband=baseband)
        else:
            self.read(inFile=datasetName, subscans=obstoProc,
                      febe=febe, baseband=baseband)

        # If chopped data, then compute phase diff
        if self.ScanParam.WobUsed:
            self.ScanParam.computeOnOff()
            self._phaseDiff()

        self.zeroStart(subscan=0)

        Plot.panels(2, 2)
        Plot.nextpage()  # start a new page
        textColBak = BogliConfig.xyouttext['color']
        BogliConfig.xyouttext['color'] = 3
        # First plot = signal before skynoise removal
        four = self.ReceiverArray.fourpixels()
        self.signal(
            four,
            noerase=1,
            caption=self.ScanParam.caption() + ' - Raw signal')
        Plot.nextpage()
        self.solveFocus(noerase=1, caption=' ')
        y1, y2 = Plot.yAxis['limits']
        Plot.xyout(0., (y1 + y2) / 2.,
                   str("%5.2f +- %4.2f" % (self.FocusOffset, self.FocusError)), size=2.5)
        Plot.plot(
            [self.FocusOffset, self.FocusOffset], y1 +
                np.array([0.7, 1.], np.float) * (y2 - y1),
                  overplot=1, style='l', ci=2)
        x1, x2 = Plot.xAxis['limits']
        Plot.xyout(
            x1 + 0.85 * (x2 - x1),
            y1 + 0.9 * (y2 - y1),
            'Raw data',
            size=2)

        # again after skynoise removal
        self.medianNoiseRemoval(computeFF=0, chanRef=-1, nbloop=3)
        Plot.nextpage()
        self.signal(four, caption='Skynoise subtracted', noerase=1)
        Plot.nextpage()
        self.solveFocus(noerase=1, caption=' ')
        y1, y2 = Plot.yAxis['limits']
        Plot.xyout(0., (y1 + y2) / 2.,
                   str("%5.2f +- %4.2f" % (self.FocusOffset, self.FocusError)), size=2.5)
        Plot.plot(
            [self.FocusOffset, self.FocusOffset], y1 +
                np.array([0.7, 1.], np.float) * (y2 - y1),
                  overplot=1, style='l', ci=2)
        modType = self.ScanParam.ScanType
        Plot.xyout(0., y1 + 0.2 * (y2 - y1), modType, size=3)
        x1, x2 = Plot.xAxis['limits']
        Plot.xyout(
            x1 + 0.82 * (x2 - x1),
            y1 + 0.91 * (y2 - y1),
            'Skynoise',
            size=2)
        Plot.xyout(
            x1 + 0.82 * (x2 - x1),
            y1 + 0.84 * (y2 - y1),
            'subtracted',
            size=2)
        Plot.panels(1, 1)
        BogliConfig.xyouttext['color'] = textColBak



# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: ReaSkydip.py (module)
DES: contains the BoA skydip class
"""
__version__ = '$Revision: 1726 $'
__date__ = '$Date: 2006-09-25 19:38:54 +0200 (Mon, 25 Sep 2006) $'

# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------

import numpy as np

from boa import ReaPointing, ReaConfig
from boa.Bogli import Plot, MultiPlot
from boa.Utilities import fitParabola, modelparabola
from boa.fortran import fStat, fFit

# ---------------------------------------------------------------------
# ---- Rea Skydip Class ------------------------------------------------
# ---------------------------------------------------------------------


class Skydip(ReaPointing.Point):

    """
    NAM: Skydip (class)
    DES: An object of this class is responsible for the reduction
         of skydips and provides zenithal opacity
    """

    def __init__(self):
        """
        DES: Initialise an instance
        """
        ReaPointing.Point.__init__(self)
        # The following two attributes are for communication with
        # apexCalibrator
        self.ResultAirmass = []
        self.ResultFluxes = []
        # Calibration V to K - this has  been measured for LABOCA
        self.VtoK = 1.E6 / 10.6
        # These parameters come into the skydip equation - default are
        # reasonable values
        self.TCabin = 283.  # +10 deg C
        self.TAtm = 270.  # -3 deg C
        self.Coupling = 0.8   # 80% optical coupling
        self.Feff = 0.97  # APEX forward efficiency at 350 GHz
        self.Tau = 0.3   # zenith opacity, that's what we want to determine

    # -------------------------------------------------------------------
    # ---- public methods -----------------------------------------------
    # -------------------------------------------------------------------
    def solveSkydip(self):
        """
        DES: compute the zenithal opacity
        """

        refChan = self.ReceiverArray.RefChannel
        flux = self.getChanData('flux', refChan)

        nSubscan = self.ScanParam.NObs
        subscanIndex = self.ScanParam.SubscanIndex

        self.MessHand.warning("Coming soon...")
        return

    def reduce(self, datasetName='', obstoProc=[], blind=317):
        """
        DES: Process a skydip scan - this method is called by the apexCalibrator
        INP: (string) datasetName: path to the dataset to be reduced
             (i list)   obstoProc: list of subscans to consider (default: all)
             (i)            blind: channel number of the ref. (blind) bolo
                               !!! SPECIFIC TO LABOCA !!!
        """
        self.read(inFile=datasetName, subscans=obstoProc)

        self.medianBaseline(subscan=0)
        # Continuous mode: try to find subscans
        if self.ScanParam.NObs == 1:
            self.findElSubscan()
        # refChan = self.ReceiverArray.RefChannel
        # for now, use channel 110
        refChan = 110
        flux = self.getChanData('flux', refChan)
        el = self.getChanData('el')

        # Number of subscans read (or found):
        nObs = len(self.ScanParam.SubscanNum)
        resultAir = []
        resultFlux = []
        for i in range(nObs):
            ind1 = self.ScanParam.SubscanIndex[0, i]
            ind2 = self.ScanParam.SubscanIndex[1, i]
            airmass = 1. / sin(el[ind1:ind2] * pi / 180.)
            resultAir.append(fStat.f_median(airmass))
            # Compute difference chan - gain*blind
            if blind:
                refFlux = self.getChanData('flux', blind)
                # Fit the relative gain
                f1 = flux[ind1:ind2]
                f2 = refFlux[ind1:ind2]
                # relG = fFit.fit_slope(f2,f1)
                # print "subscan %i - relG = %f"%(i,relG)
                # ratio = fStat.f_median(relG)
                # self.MessHand.info('Computed relative gain %i/%i =
                # %f'%(blind,refChan,ratio))
                ratio = 0.7
                diff = f1 - ratio * f2
            else:
                refFlux = 0
                diff = flux[ind1:ind2] - flux[0]

            resultFlux.append(fStat.f_median(diff))
            # resultFlux.append(fStat.f_median(diff[ind1:ind2]))

        self.ResultAirmass = resultAir
        self.ResultFluxes = resultFlux

    def findElAccSubscan(self):
        """
        determine subscans by looking at elevation acceleration
        """
        ae = self.getChanData('elacc', flag='None')
                              # get ALL in order to compute subscan indices
        ind1 = [0]
        ind2 = []
        nb = len(ae)
        i = 0
        while i < nb - 3:
            while((ae[i] > -2000 or ae[i + 1] > -2000 or ae[i + 2] > -2000) and i < nb - 3):
                i += 1
            ind2.append(i)
            while((ae[i] < 2000 or ae[i + 1] < 2000 or ae[i + 2] < 2000) and i < nb - 3):
                i += 1
            while((ae[i] > 2000 or ae[i + 1] > 2000 or ae[i + 2] > 2000) and i < nb - 3):
                i += 1
            ind1.append(i)
        ind1 = ind1[:-1]

        # Reinitialise Subscan related attributes
        self.ScanParam.SubscanIndex = []
        self.ScanParam.SubscanNum = []
        self.ScanParam.SubscanType = []
        nb = len(ind1)
        for i in range(nb):
            self.ScanParam.SubscanIndex.append([ind1[i], ind2[i]])
            self.ScanParam.SubscanNum.append(i + 1)
            self.ScanParam.SubscanType.append('ON')

        self.ScanParam.SubscanIndex = np.transpose(self.ScanParam.SubscanIndex)
        self.ScanParam.NObs = nb
        self.MessHand.info(str("Found %i subscans" % (nb)))

    def findElSubscan(self, tolerance=0.2, delta=5.e-4):
        """
        DES: Determine subscans by looking at elevation
        INP: (f) tolerance: tolerance (in deg) to assume that we're still in the same subscan
             (f)    delta : minimum variation of El (in deg) for the steps between subscans
        """
        el = self.getChanData('el')
        dEl = abs(el[1:] - el[:-1])
        ind1 = []
        ind2 = []
        i = 0
        nb = len(dEl)
        while i < nb - 1:
            # The following is true when between two subscans
            while dEl[i] > delta and i < nb - 1:
                i += 1
            currEl = el[i]
            ind1.append(i)
            # The following stays true during one subscan
            while abs(el[i] - currEl) < tolerance and i < nb - 1:
                i += 1
            ind2.append(i)

        # Reinitialise Subscan related attributes
        self.ScanParam.SubscanIndex = []
        self.ScanParam.SubscanNum = []
        self.ScanParam.SubscanType = []
        nb = len(ind1)
        for i in range(nb):
            self.ScanParam.SubscanIndex.append([ind1[i], ind2[i]])
            self.ScanParam.SubscanNum.append(i + 1)
            self.ScanParam.SubscanType.append('ON')

        self.ScanParam.SubscanIndex = np.transpose(self.ScanParam.SubscanIndex)
        self.ScanParam.NObs = nb
        self.MessHand.info(str("Found %i subscans" % (nb)))



# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
..modules:: ReaMessageHandler
:synopsis: contains the Rea message handler class
"""
__version__ = '$Revision: 2195 $'
__date__ = '$Date: 2007-07-15 16:45:11 +0200 (dim, 15 jui 2007) $'

#----------------------------------------------------------------------------------
#----- Import ---------------------------------------------------------------------
#-------------------------------------------------------------------------
import os
import sys
import string
import time
from rea import ReaConfig

# Keeps all the messages in memory
_history = {}  # for test


def printHistory():
    allkeys = sorted(_history.keys())

    for key in allkeys:
        print key, _history[key]

#----------------------------------------------------------------------------------
#----- Rea Message Handler --------------------------------------------------------
#-------------------------------------------------------------------------


class MessHand:

    """..class:: MessHand
    :synopsis: An object of this class is responsible for the management of output
    messages as well as the creation of message files.
    """

    def __init__(self, logName='Unknown'):
        """initialise an instance

        Parameters
        ----------
        logName : str
            the filename of the logfile
        """

        # parameter attributes:

        # default value for max. weight of messages to be printed
        self.maxWeight = ReaConfig.maxMessHandWeight

        # Private attributes:

        self.__logName = ''  # Name of the class using the Message Handler
                            # 20 char long left hand stripped.
        self.__WeightList = {
            1: 'E:',
            2: 'W:',
            3: 'I:',
            4: 'L:',
            5: 'D:'}  # list of allowed weights
        self.__WeightDescription = ['errors', 'warnings',
                                    'short info', 'extended info',
                                    'debug']

        self.__messFileName = "rea.mes"                                   # default name of message file
# self.__prompt = chr(1) + "\033[1;32m" + chr(2) + 'rea>' + chr(1) +
# "\033[0m" + chr(2)
        self.__prompt = 'rea<'

        # methods to run at initialisation
        self.setLogName(logName)

    def __del__(self):
        self.debug("closing message handler '" + self.__logName.strip() + "'")

    #--------------------------------------------------------------------------------
    #----- methods ------------------------------------------------------------------
    #-------------------------------------------------------------------------
    def Welcome(self):
        """print welcome message"""
        str = ""
        print
        print " Welcome to The REceiver Array data Analysis project ! "
        print "   -------------------------------------------    "

        if os.environ.get('USER'):
            str += " User: " + os.environ.get('USER') + " "
        str += "(" + time.ctime() + ")\n"
        #   str += "Running on " +sys.platform+" with Python "+sys.version
        print str

    #-------------------------------------------------------------------------
    def setMaxWeight(self, weight='2'):
        """Set the maximum weight of messages to be printed.

        Parameters
        ----------
        weight : int
            maximum weight
                      1: errors, queries
                      2: warnings
                      3: short info
                      4: extended info
                      5: debug
        """
        if weight in self.__WeightList:
            self.maxWeight = weight
            self.info("max weight of messages = " + repr(self.maxWeight))
        else:
            self.error("invalid max weight of messages: " + repr(weight))

    #-------------------------------------------------------------------------
    def setLogName(self, logName='Unknown'):
        self.__logName = logName.ljust(20)    # Name of the class using
        self.__logName = self.__logName[0:20]  # the Message Handler 20 char
                                              # long left hand stripped

    #-------------------------------------------------------------------------
    def pause(self, message=''):
        """allow to make a pause in the program

        Parameters
        ----------
        message : str
            a message to display
        """
        if not message:
            message = 'Please press enter to continue'

        self.ask(message)

    #-------------------------------------------------------------------------
    def yesno(self, message=''):
        """ask the user a question with yes/no answer type

        Parameters
        ----------
        message : str
            the question

        Returns
        -------
        bool
            the answer
        """

        yes = ['y', 'Y', 'yes', 'Yes', 'YES']
        no = ['n', 'N', 'no', 'No', 'NO']
        answer = ''
        while answer not in yes and answer not in no:
            answer = self.ask(message + ' (y/n): ')

        if answer in yes:
            return 1
        else:
            return 0

    #-------------------------------------------------------------------------
    def ask(self, message=''):
        """ask the user

        Parameters
        ----------
        message : str
            the question

        Returns
        -------
        str
            the answer
        """
        return raw_input(self.__prompt + ' ?: ' + message)

    #-------------------------------------------------------------------------
    def error(self, message=''):
        """to print an error message

        Parameters
        ----------
        message : str
            the error message
        """
        self.setMess(1, message)

    #-------------------------------------------------------------------------
    def warning(self, message=''):
        """to print an warning message

        Parameters
        ----------
        message : str
            the warning message
        """
        self.setMess(2, message)

    #-------------------------------------------------------------------------
    def info(self, message=''):
        """to print an info message

        Parameters
        ----------
        message : str
            the info message
        """
        self.setMess(3, message)

    #-------------------------------------------------------------------------
    def longinfo(self, message=''):
        """to print an long info message

        Parameters
        ----------
        message : str
            the long information message
        """
        self.setMess(4, message)

    #-------------------------------------------------------------------------
    def debug(self, message=''):
        """to print an debug message

        Parameters
        ----------
        message : str
            the debug message
        """
        self.setMess(5, message)

    #-------------------------------------------------------------------------
    def line(self):
        """to print a line"""
        message = '-' * (80 - 10)
        self.setMess(4, message)

    #-------------------------------------------------------------------------
    def setMess(self, weight=1, message=' '):
        """deposit messages for screen output and message files

        Parameters
        ----------
        weight : int
            weight of transferred message (see setMaxWeight)
              1 - error
              2 - warning
              3 - info
              4 - longinfo
              5 - debug
        message : str
            message to be printed and added to message file
        """
        if weight in self.__WeightList:
            message = message.strip()
            short_prefix = " " + self.__WeightList[weight] + " "
            long_prefix = time.ctime() + " " + self.__logName + short_prefix

            if weight <= self.maxWeight:                         # Print if asked ...

                # ... in file
                if (self.__messFileName != ""):
                    messFile = open(self.__messFileName, 'a')
                    for line in message.split('\n'):
                        messFile.write(long_prefix + line + "\n")
                    messFile.close()
                    del messFile
                else:

                    print self.__prompt + " " + self.__WeightList[1] + " " + \
                        "the log file for this message handler is closed or undefined"

                # ... on screen
                for line in message.split('\n'):
                    print self.__prompt + short_prefix + line

                # and add this to history
                _history[long_prefix] = message

                del short_prefix, long_prefix

    #-------------------------------------------------------------------------
    def initMessFile(self, filename="rea.mes"):
        """set & initialise new message file

        Parameters
        ----------
        filename : str
            log filename
        """

        if self.__messFileName == "":
            self.__messFileName = os.path.join(
                ReaConfig.outDir, filename.strip())

        try:

            if os.path.isfile(self.__messFileName):
                # File name already exist so move it to a new location
                # First find a available file name as self.__messFileName+i
                i = 1
                while(os.path.isfile(self.__messFileName + str(i))):
                    i += 1
                newFileName = self.__messFileName + str(i)
                self.info('old log file renamed to ' + newFileName)
                os.rename(self.__messFileName, newFileName)

            messFile = open(self.__messFileName, 'w')
            messFile.write(
                "Logfile for The REceiver Array data Analysis project. ")
            messFile.write("created on " + time.ctime() + "\n")
            messFile.close()
            del messFile

        except IOError:
            self.error("cannot open message file " + self.__messFileName)

    #-------------------------------------------------------------------------
    def closeMessFile(self):
        """set self.__existMessFile to 0 and file name to ""
        """
        self.debug("closing message file " + self.__messFileName)
        self.__messFileName = ""

#----------------------------------------------------------------------------------
#----- For compatibility with the CalibratorLog class -----------------------------
#-------------------------------------------------------------------------


class Logger:

    """..class:: Logger
    :synopsis: for compatiliby with the CalibratorLog.Logger class
    """

    def __init__(self, logType='ACS'):
        """Initiabise an instance"""

        self.logger = printLogger()


class printLogger(MessHand):

    """..class:: printLogger
    :synopsis: for compatibility with the CalibratorLog.printLogger class
    """

    def __init__(self):

        MessHand.__init__(self)
        self.setLogName(logName='CalibratorLog')

    #--------------------------------------------------------------------------------
    # Compatibility with the CalibratorLog class
    def logInfo(self, message):
        self.info(message)

    def logError(self, message):
        self.error(message)

    def logWarning(self, message):
        self.warning(message)

    def logDebug(self, message):
        self.debug(message)
